import shutil
import warnings
from contextlib import contextmanager
from pathlib import Path

import pytest

from pharmpy.internals.fs.cwd import chdir
from pharmpy.model import Model
from pharmpy.tools import run_amd
from pharmpy.tools.amd.run import validate_input
from pharmpy.workflows import default_tool_database


def test_invalid_search_space_raises(tmp_path, testdata):
    with chdir(tmp_path):
        db, model = _load_model(testdata)

        with pytest.raises(
            ValueError,
            match='Invalid `search_space`, could not be parsed:',
        ):
            run_amd(
                model,
                results=model.modelfit_results,
                search_space='XYZ',
                retries_strategy="skip",
                path=db.path,
                resume=True,
            )


@pytest.mark.filterwarnings(
    'ignore::UserWarning',
)
def test_skip_most(tmp_path, testdata):
    with chdir(tmp_path):
        db, model = _load_model(testdata)

        to_be_skipped = validate_input(
            model,
            results=model.modelfit_results,
            modeltype='basic_pk',
            administration='oral',
            occasion=None,
            retries_strategy="skip",
            path=db.path,
            resume=True,
        )

    assert len(to_be_skipped) == 3


@pytest.mark.filterwarnings(
    'ignore::UserWarning',
)
def test_raise_allometry(tmp_path, testdata):
    with chdir(tmp_path):
        db, model = _load_model(testdata, with_datainfo=True)

        with pytest.raises(
            ValueError,
            match='Invalid `allometric_variable`',
        ):
            run_amd(
                model,
                results=model.modelfit_results,
                modeltype='basic_pk',
                administration='oral',
                allometric_variable='SJDLKSDJ',
                retries_strategy="skip",
                path=db.path,
                resume=True,
            )


@pytest.mark.filterwarnings(
    'ignore::UserWarning',
)
def test_raise_covsearch(tmp_path, testdata):
    with chdir(tmp_path):
        db, model = _load_model(testdata, with_datainfo=True)

        with pytest.raises(
            ValueError,
            match='Invalid `search_space` because of invalid covariate .* got `SJDLKSDJ`',
        ):
            run_amd(
                model,
                results=model.modelfit_results,
                search_space='LET(CONTINUOUS, [AGE, SJDLKSDJ]); LET(CATEGORICAL, [SEX])',
                modeltype='basic_pk',
                administration='oral',
                retries_strategy="skip",
                path=db.path,
                resume=True,
            )


@pytest.mark.filterwarnings(
    'ignore::UserWarning',
)
def test_skip_covsearch(tmp_path, testdata):
    with chdir(tmp_path):
        db, model = _load_model(testdata)
        to_be_skipped = validate_input(
            model,
            results=model.modelfit_results,
            search_space='LET(CONTINUOUS, []); LET(CATEGORICAL, [])',
            modeltype='basic_pk',
            administration='oral',
            occasion='VISI',
            allometric_variable='WT',
            retries_strategy="skip",
            path=db.path,
            resume=True,
        )
    assert len(to_be_skipped) == 1


@pytest.mark.filterwarnings(
    'ignore::UserWarning',
)
def test_skip_iovsearch_one_occasion(tmp_path, testdata):
    with chdir(tmp_path):
        db, model = _load_model(testdata, with_datainfo=True)

        to_be_skipped = validate_input(
            model,
            results=model.modelfit_results,
            modeltype='basic_pk',
            administration='oral',
            retries_strategy="skip",
            occasion='XAT2',
            path=db.path,
            resume=True,
        )

    assert len(to_be_skipped) == 1


@pytest.mark.filterwarnings(
    'ignore::UserWarning',
)
def test_skip_iovsearch_missing_occasion_raises(tmp_path, testdata):
    with chdir(tmp_path):
        db, model = _load_model(testdata)

        with pytest.raises(
            ValueError,
            match='Invalid `occasion`',
        ):
            run_amd(
                model,
                results=model.modelfit_results,
                modeltype='basic_pk',
                administration='oral',
                occasion='XYZ',
                retries_strategy="skip",
                path=db.path,
                resume=True,
            )


@pytest.mark.filterwarnings(
    'ignore::UserWarning',
)
def test_ignore_datainfo_fallback(tmp_path, testdata):
    with chdir(tmp_path):
        db, model = _load_model(testdata, with_datainfo=True)

        to_be_skipped = validate_input(
            model,
            results=model.modelfit_results,
            modeltype='basic_pk',
            administration='oral',
            retries_strategy="skip",
            ignore_datainfo_fallback=True,
            path=db.path,
            resume=True,
        )

    assert len(to_be_skipped) == 3


def _load_model(testdata: Path, with_datainfo: bool = False):
    models = testdata / 'nonmem' / 'models'

    # NOTE: We need to make a local copy and read the model locally to avoid
    # reading the .datainfo which contains allometry information we do not want
    # to ignore.

    shutil.copy2(models / 'mox_simulated_normal.csv', '.')
    if with_datainfo:
        shutil.copy2(models / 'mox_simulated_normal.datainfo', '.')
    shutil.copy2(models / 'mox2.mod', '.')
    shutil.copy2(models / 'mox2.ext', '.')
    shutil.copy2(models / 'mox2.lst', '.')
    shutil.copy2(models / 'mox2.phi', '.')

    model = Model.parse_model('mox2.mod')
    model = model.replace(name='start')

    # NOTE: Load results directly in DB to skip fitting
    db_tool = default_tool_database(toolname='amd', path='amd_dir1')
    db_fit = default_tool_database(toolname='modelfit', path=db_tool.path / 'modelfit')

    with db_fit.model_database.transaction(model) as txn:
        txn.store_model()
        txn.store_modelfit_results()
        # NOTE: This are needed because currently caching of the results cannot
        # read from the JSON file created above.
        txn.store_local_file(models / 'mox2.ext', 'start.ext')
        txn.store_local_file(models / 'mox2.lst', 'start.lst')
        txn.store_local_file(models / 'mox2.phi', 'start.phi')

    return db_tool, model


@contextmanager
def _record_warnings():
    with pytest.warns(Warning) as record:
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                module='distributed',
                category=UserWarning,
            )
            warnings.filterwarnings(
                "ignore",
                module='distributed',
                category=ResourceWarning,
            )
            warnings.filterwarnings(
                "ignore",
                module='distributed',
                category=RuntimeWarning,
            )
            yield record


def _validate_record(record, expected):
    assert len(record) == len(expected)

    for warning, match in zip(record, expected):
        assert match in str(warning.message)



import numpy as np
import pandas as pd
from pytest import approx

import pharmpy.tools.cdd.results as cdd
from pharmpy.tools import read_modelfit_results
from pharmpy.tools.psn_helpers import model_paths, options_from_command


def test_psn_cdd_options(testdata):
    opts = cdd.psn_cdd_options(testdata / 'nonmem' / 'cdd' / 'pheno_real_bin10')
    assert opts['case_column'] == 'ID'
    assert (
        opts['model_path']
        == '/home/PMX/Projects/github/pharmpy/tests/testdata/nonmem/pheno_real.mod'
    )


def test_psn_options():
    cmd = 'cmd_line: /opt/bin/cdd pheno.mod -ignore -bins=10 -case=WGT -clean=3 -dir=caseWGTbin10'
    assert options_from_command(cmd) == {
        'ignore': '',
        'bins': '10',
        'case': 'WGT',
        'clean': '3',
        'dir': 'caseWGTbin10',
    }


def test_cdd_psn(load_model_for_test, testdata):
    path = testdata / 'nonmem' / 'cdd' / 'pheno_real_bin10'
    base_model = load_model_for_test(testdata / 'nonmem' / 'cdd' / 'pheno_real.mod')
    base_model_results = read_modelfit_results(testdata / 'nonmem' / 'cdd' / 'pheno_real.mod')
    cdd_models = [load_model_for_test(p) for p in model_paths(path, 'cdd_*.mod')]
    cdd_model_results = [read_modelfit_results(p) for p in model_paths(path, 'cdd_*.mod')]
    skipped_individuals = cdd.psn_cdd_skipped_individuals(path)

    cdd_bin_id = cdd.calculate_results(
        base_model, base_model_results, cdd_models, cdd_model_results, 'ID', skipped_individuals
    )

    correct = pd.DataFrame(
        {
            'cook_score': [
                0.6002247,
                0.7681936,
                1.054763,
                0.8994126,
                0.9405288,
                0.6945769,
                0.7631572,
                0.7947513,
                0.7535628,
                0.5651823,
            ],
            'jackknife_cook_score': [
                0.55860759,
                0.90000807,
                1.00565998,
                0.92285602,
                0.74943641,
                0.83529442,
                0.89191965,
                0.85322234,
                0.6674856,
                0.70395028,
            ],
            'delta_ofv': [
                0.27067,
                0.35146,
                0.79115,
                0.84426,
                0.86966,
                0.37618,
                0.69805,
                0.49343,
                0.31184,
                0.31729,
            ],
            'dofv_influential': [
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
            ],
            'covariance_ratio': [
                1.218597,
                1.389662,
                0.906991,
                2.152456,
                1.166288,
                1.214472,
                1.314591,
                1.085536,
                1.337584,
                1.342893,
            ],
            'skipped_individuals': [
                ['1', '2', '3', '4', '5', '6'],
                ['7', '8', '9', '10', '11', '12'],
                ['13', '14', '15', '16', '17', '18'],
                ['19', '20', '21', '22', '23', '24'],
                ['25', '26', '27', '28', '29', '30'],
                ['31', '32', '33', '34', '35', '36'],
                ['37', '38', '39', '40', '41', '42'],
                ['43', '44', '45', '46', '47', '48'],
                ['49', '50', '51', '52', '53', '54'],
                ['55', '56', '57', '58', '59'],
            ],
        },
        index=[
            1,
            2,
            3,
            4,
            5,
            6,
            7,
            8,
            9,
            10,
        ],
    )
    pd.testing.assert_frame_equal(cdd_bin_id.case_results, correct, rtol=1e-4)


def test_cdd_calculate_results(load_model_for_test, testdata):
    path = testdata / 'nonmem' / 'cdd' / 'pheno_real_bin10'
    skipped_individuals = cdd.psn_cdd_skipped_individuals(path)
    base_model = load_model_for_test(testdata / 'nonmem' / 'cdd' / 'pheno_real.mod')
    base_model_results = read_modelfit_results(testdata / 'nonmem' / 'cdd' / 'pheno_real.mod')
    cdd_model_paths = model_paths(path, 'cdd_*.mod')

    cdd_models = [load_model_for_test(p) for p in cdd_model_paths]
    cdd_model_results = [read_modelfit_results(p) for p in cdd_model_paths]

    # Results for plain PsN run
    delta_ofv = cdd.compute_delta_ofv(base_model_results, cdd_model_results, skipped_individuals)

    assert delta_ofv == approx(
        [0.27067, 0.35146, 0.79115, 0.84426, 0.86966, 0.37618, 0.69805, 0.49343, 0.31184, 0.31729],
        abs=1e-5,
    )

    cdd_estimates = pd.DataFrame(
        data=[
            pd.Series(res.parameter_estimates, name=m.name)
            for m, res in zip(cdd_models, cdd_model_results)
            if res is not None
        ]
    )

    jackknife = cdd.compute_jackknife_covariance_matrix(cdd_estimates)
    jackpsnvalues = [
        3.30415232e-08,
        -1.02486433e-06,
        -3.42829379e-07,
        -6.31160018e-07,
        -6.54383718e-08,
        4.16059589e-08,
        -1.02486433e-06,
        7.66862781e-04,
        -8.00400982e-04,
        7.94700855e-05,
        7.30565442e-06,
        4.83776339e-06,
        -3.42829379e-07,
        -8.00400982e-04,
        8.97549534e-03,
        -2.91263334e-04,
        2.17285280e-04,
        3.56099773e-05,
        -6.31160018e-07,
        7.94700855e-05,
        -2.91263334e-04,
        1.62624661e-04,
        -5.66902777e-05,
        3.77862212e-06,
        -6.54383718e-08,
        7.30565442e-06,
        2.17285280e-04,
        -5.66902777e-05,
        5.45538028e-05,
        -2.00496073e-06,
        4.16059589e-08,
        4.83776339e-06,
        3.56099773e-05,
        3.77862212e-06,
        -2.00496073e-06,
        2.62509124e-06,
    ]

    assert jackknife.values.flatten() == approx(jackpsnvalues, rel=1e-6)

    cook_scores = cdd.compute_cook_scores(
        base_model_results.parameter_estimates,
        cdd_estimates,
        base_model_results.covariance_matrix,
    )

    assert cook_scores == approx(
        [
            0.6002247,
            0.7681936,
            1.054763,
            0.8994126,
            0.9405288,
            0.6945769,
            0.7631572,
            0.7947513,
            0.7535628,
            0.5651823,
        ],
        abs=1e-5,
    )

    cov_ratios = cdd.compute_covariance_ratios(
        cdd_model_results, base_model_results.covariance_matrix
    )

    assert cov_ratios == approx(
        [
            1.218597,
            1.389662,
            0.906991,
            2.152456,
            1.166288,
            1.214472,
            1.314591,
            1.085536,
            1.337584,
            1.342893,
        ],
        abs=1e-5,
    )

    # Replace three estimated cdd_models with fake models without estimates
    # and recompute results to verify handling of missing output
    cdd_models[0] = load_model_for_test(path / 'm1' / 'rem_1.mod')
    cdd_model_results[0] = read_modelfit_results(path / 'm1' / 'rem_1.mod')
    cdd_models[1] = load_model_for_test(path / 'm1' / 'rem_2.mod')
    cdd_model_results[1] = read_modelfit_results(path / 'm1' / 'rem_2.mod')
    cdd_models[3] = load_model_for_test(path / 'm1' / 'rem_4.mod')
    cdd_model_results[3] = read_modelfit_results(path / 'm1' / 'rem_4.mod')

    res = cdd.calculate_results(
        base_model, base_model_results, cdd_models, cdd_model_results, 'ID', skipped_individuals
    )

    assert res.case_results.delta_ofv.values == approx(
        [np.nan, np.nan, 0.79115, np.nan, 0.86966, 0.37618, 0.69805, 0.49343, 0.31184, 0.31729],
        abs=1e-5,
        nan_ok=True,
    )

    # pandas.testing.assert_frame_equal  rtol atol

    assert res.case_results.covariance_ratio.values == approx(
        [
            np.nan,
            np.nan,
            0.906991,
            np.nan,
            1.166288,
            1.214472,
            1.314591,
            1.085536,
            1.337584,
            1.342893,
        ],
        abs=1e-5,
        nan_ok=True,
    )

    assert res.case_results.cook_score.values == approx(
        [
            np.nan,
            np.nan,
            1.054763,
            np.nan,
            0.9405288,
            0.6945769,
            0.7631572,
            0.7947513,
            0.7535628,
            0.5651823,
        ],
        abs=1e-5,
        nan_ok=True,
    )


# tox -e py38 -- pytest -s test_parameter.py



import datetime

from pharmpy.deps import pandas as pd
from pharmpy.internals.immutable import Immutable

CATEGORIES = ('ERROR', 'WARNING', 'INFORMATION')


class LogEntry(Immutable):
    def __init__(self, category: str, message: str, time: datetime.datetime):
        self._category = category
        self._message = message
        self._time = time

    @classmethod
    def create(cls, category: str, message: str):
        if category not in CATEGORIES:
            raise ValueError(f"Unknown category {category}")
        return cls(category=category, message=message, time=datetime.datetime.now())

    @property
    def category(self):
        """Category of message. One of ERROR, WARNING and INFORMATION"""
        return self._category

    @property
    def message(self):
        """The logged message"""
        return self._message

    @property
    def time(self):
        """Time stamp of log entry"""
        return self._time

    def to_dict(self):
        return {
            'category': self._category,
            'message': self._message,
            'time': self._time.isoformat(),
        }

    @classmethod
    def from_dict(cls, d):
        d['time'] = datetime.datetime.fromisoformat(d['time'])
        return LogEntry(**d)


class Log:
    """Timestamped error and warning log"""

    def __init__(self, entries: tuple[LogEntry, ...] = ()):
        self._entries = entries

    def __len__(self):
        return len(self._entries)

    def __iter__(self):
        return iter(self._entries)

    def to_dict(self):
        return {i: entry.to_dict() for i, entry in enumerate(self._entries)}

    @classmethod
    def from_dict(cls, d):
        entries = []
        for entry in d.values():
            log_entry = LogEntry.from_dict(entry)
            entries.append(log_entry)
        return cls(tuple(entries))

    @property
    def errors(self):
        entries = [e for e in self._entries if e.category == 'ERROR']
        return Log(tuple(entries))

    @property
    def warnings(self):
        entries = [e for e in self._entries if e.category == 'WARNING']
        return Log(tuple(entries))

    def log_error(self, message):
        """Log an error

        Parameters
        ----------
        message : str
            Error message
        """
        entry = LogEntry.create(category='ERROR', message=message)
        entries = self._entries + (entry,)
        return Log(entries)

    def log_warning(self, message):
        """Log a warning

        Parameters
        ----------
        message : str
            Warning message
        """
        entry = LogEntry.create(category='WARNING', message=message)
        entries = self._entries + (entry,)
        return Log(entries)

    def to_dataframe(self):
        """Create an overview dataframe from log

        Returns
        -------
        pd.DataFrame
            Dataframe with overview of log entries
        """
        categories = [entry.category for entry in self._entries]
        times = [entry.time for entry in self._entries]
        messages = [entry.message for entry in self._entries]
        df = pd.DataFrame(
            {
                'category': categories,
                'time': times,
                'message': messages,
            }
        )
        return df

    def __repr__(self):
        s = ''
        for entry in self._entries:
            s += f'{entry.category:<8} {entry.time.strftime("%Y-%m-%d %H:%M:%S")}\n'
            for line in entry.message.split('\n'):
                s += f'    {line}\n'
        return s



from __future__ import annotations

from collections.abc import Sequence as CollectionsSequence
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Union,
    overload,
)

from pharmpy.internals.immutable import Immutable, cache_method

if TYPE_CHECKING:
    import numpy as np
    import pandas as pd
    import sympy
else:
    from pharmpy.deps import numpy as np
    from pharmpy.deps import pandas as pd
    from pharmpy.deps import sympy


class Parameter(Immutable):
    """A single parameter

    Example
    -------

    >>> from pharmpy.model import Parameter
    >>> param = Parameter("TVCL", 0.005, lower=0.0)
    >>> param.init
    0.005

    Parameters
    ----------
    name : str
        Name of the parameter
    init : number
        Initial estimate or simply the value of parameter.
    fix : bool
        A boolean to indicate whether the parameter is fixed or not. Note that fixing a parameter
        will keep its bounds even if a fixed parameter is actually constrained to one single
        value. This is so that unfixing will take back the previous bounds.
    lower : float
        The lower bound of the parameter. Default no bound. Must be less than the init.
    upper : float
        The upper bound of the parameter. Default no bound. Must be greater than the init.
    """

    def __init__(
        self,
        name: str,
        init: float,
        lower: float = -float("inf"),
        upper: float = float("inf"),
        fix: bool = False,
    ):
        self._name = name
        self._init = init
        self._lower = lower
        self._upper = upper
        self._fix = fix

    @classmethod
    def create(
        cls,
        name: str,
        init: Union[float, sympy.Float],
        lower: Optional[Union[float, sympy.Float]] = None,
        upper: Optional[Union[float, sympy.Float]] = None,
        fix: bool = False,
    ):
        """Alternative constructor for Parameter with error checking"""
        if not isinstance(name, str):
            raise ValueError("Name of parameter must be of type string")
        init = float(init)
        if np.isnan(init):
            raise ValueError('Initial estimate cannot be NaN')
        if lower is None:
            lower = -float('inf')
        else:
            lower = float(lower)
        if upper is None:
            upper = float('inf')
        else:
            upper = float(upper)
        if init < lower:
            raise ValueError(f'Lower bound {lower} cannot be greater than init {init}')
        if init > upper:
            raise ValueError(f'Upper bound {upper} cannot be less than init {init}')
        return cls(name, init, lower, upper, bool(fix))

    def replace(self, **kwargs):
        """Replace properties and create a new Parameter"""
        name = kwargs.get('name', self._name)
        init = kwargs.get('init', self._init)
        lower = kwargs.get('lower', self._lower)
        upper = kwargs.get('upper', self._upper)
        fix = kwargs.get('fix', self._fix)
        new = Parameter.create(name, init, lower=lower, upper=upper, fix=fix)
        return new

    @property
    def name(self) -> str:
        """Parameter name"""
        return self._name

    @property
    def fix(self) -> bool:
        """Should parameter be fixed or not"""
        return self._fix

    @property
    def symbol(self) -> sympy.Symbol:
        """Symbol representing the parameter"""
        return sympy.Symbol(self._name)

    @property
    def lower(self) -> float:
        """Lower bound of the parameter"""
        return self._lower

    @property
    def upper(self) -> float:
        """Upper bound of the parameter"""
        return self._upper

    @property
    def init(self) -> float:
        """Initial parameter estimate or value"""
        return self._init

    @cache_method
    def __hash__(self):
        return hash((self.name, self.init, self.lower, self.upper, self.fix))

    def to_dict(self) -> Dict[str, Any]:
        return {
            'name': self.name,
            'init': self.init,
            'lower': self.lower,
            'upper': self.upper,
            'fix': self.fix,
        }

    @classmethod
    def from_dict(cls, d: Dict[str, Any]):
        return cls(**d)

    def __eq__(self, other):
        """Two parameters are equal if they have the same name, init and constraints"""
        if hash(self) != hash(other):
            return False
        return (
            isinstance(other, Parameter)
            and self._init == other._init
            and self._lower == other._lower
            and self._upper == other._upper
            and self._name == other._name
            and self._fix == other._fix
        )

    def __repr__(self):
        if self._lower == -float("inf"):
            lower = "-∞"
        else:
            lower = self._lower
        if self._upper == float("inf"):
            upper = "∞"
        else:
            upper = self._upper
        return (
            f'Parameter("{self._name}", {self._init}, lower={lower}, upper={upper}, '
            f'fix={self._fix})'
        )


class Parameters(CollectionsSequence, Immutable):
    """An immutable collection of parameters

    Class representing a group of parameters. Usually all parameters in a model.
    This class give a ways of displaying, summarizing and manipulating
    more than one parameter at a time.

    Specific parameters can be found using indexing on the parameter name

    Example
    -------

    >>> from pharmpy.model import Parameters, Parameter
    >>> par1 = Parameter("x", 0)
    >>> par2 = Parameter("y", 1)
    >>> pset = Parameters((par1, par2))
    >>> pset["x"]
    Parameter("x", 0, lower=-∞, upper=∞, fix=False)

    >>> "x" in pset
    True

    """

    def __init__(self, parameters: Tuple[Parameter, ...] = ()):
        self._params = parameters

    @classmethod
    def create(cls, parameters: Optional[Union[Parameters, Sequence[Parameter]]] = None):
        if isinstance(parameters, Parameters):
            return parameters
        elif parameters is None:
            parameters = ()
        else:
            parameters = tuple(parameters)
        names = set()
        for p in parameters:
            if not isinstance(p, Parameter):
                raise ValueError(f'Can not add variable of type {type(p)} to Parameters')
            if p.name in names:
                raise ValueError(
                    f'Parameter names must be unique. Parameter "{p.name}" '
                    'was added more than once to Parameters'
                )
            names.add(p.name)
        return cls(parameters)

    def replace(self, **kwargs):
        """Replace properties and create a new Parameters object"""
        parameters = kwargs.get('parameters', self._params)
        new = Parameters.create(parameters)
        return new

    def __len__(self):
        return len(self._params)

    def _lookup_param(self, ind: Union[int, str, sympy.Symbol, Parameter]):
        if isinstance(ind, sympy.Symbol):
            ind = ind.name
        if isinstance(ind, str):
            for i, param in enumerate(self._params):
                if ind == param.name:
                    return i, param
            raise KeyError(f'Could not find {ind} in Parameters')
        elif isinstance(ind, Parameter):
            try:
                i = self._params.index(ind)
            except ValueError:
                raise KeyError(f'Could not find {ind.name} in Parameters')
            return i, ind
        return ind, self._params[ind]

    @overload
    def __getitem__(self, ind: Union[int, str, sympy.Symbol, Parameter]) -> Parameter:
        ...

    @overload
    def __getitem__(
        self, ind: Union[slice, Sequence[Union[int, str, sympy.Symbol, Parameter]]]
    ) -> Parameters:
        ...

    def __getitem__(self, ind):
        if isinstance(ind, slice):
            return Parameters(self._params[ind])
        elif not isinstance(ind, str) and isinstance(ind, CollectionsSequence):
            return Parameters(tuple(self._lookup_param(i)[1] for i in ind))
        else:
            _, param = self._lookup_param(ind)
            return param

    def __contains__(self, ind):
        try:
            self._lookup_param(ind)
            return True
        except KeyError:
            return False

    def to_dataframe(self) -> pd.DataFrame:
        """Create a dataframe with a summary of all Parameters

        Returns
        -------
        DataFrame
            A dataframe with one row per parameter.
            The columns are value, lower, upper and fix
            Row Index is the names

        Example
        -------
        >>> from pharmpy.model import Parameters, Parameter
        >>> par1 = Parameter("CL", 1, lower=0, upper=10)
        >>> par2 = Parameter("V", 10, lower=0, upper=100)
        >>> pset = Parameters((par1, par2))
        >>> pset.to_dataframe()
            value  lower  upper    fix
        CL      1      0     10  False
        V      10      0    100  False

        """
        symbols = [param.name for param in self._params]
        values = [param.init for param in self._params]
        lower = [param.lower for param in self._params]
        upper = [param.upper for param in self._params]
        fix = [param.fix for param in self._params]
        return pd.DataFrame(
            {'value': values, 'lower': lower, 'upper': upper, 'fix': fix}, index=symbols
        )

    @property
    def names(self) -> List[str]:
        """List of all parameter names"""
        return [p.name for p in self._params]

    @property
    def symbols(self) -> List[sympy.Symbol]:
        """List of all parameter symbols"""
        return [p.symbol for p in self._params]

    @property
    def lower(self) -> Dict[str, float]:
        """Lower bounds of all parameters as a dictionary"""
        return {p.name: p.lower for p in self._params}

    @property
    def upper(self) -> Dict[str, float]:
        """Upper bounds of all parameters as a dictionary"""
        return {p.name: p.upper for p in self._params}

    @property
    def inits(self) -> Dict[str, float]:
        """Initial estimates of parameters as dict"""
        return {p.name: p.init for p in self._params}

    def set_initial_estimates(self, inits: Mapping[str, float]):
        """Create a new Parameters with changed initial estimates

        Parameters
        ----------
        inits : dict
            A dictionary of parameter names to initial estimates

        Return
        ------
        Parameters
            An update Parameters object
        """
        new = []
        for p in self:
            if p.name in inits:
                newparam = p.replace(init=inits[p.name])
            else:
                newparam = p
            new.append(newparam)
        return Parameters(tuple(new))

    @property
    def fix(self) -> Dict[str, bool]:
        """Fixedness of parameters as dict"""
        return {p.name: p.fix for p in self._params}

    def set_fix(self, fix: Mapping[str, bool]):
        """Create a new Parameters with changed fix state

        Parameters
        ----------
        fix : dict
            A dictionary of parameter names to boolean fix state

        Return
        ------
        Parameters
            An update Parameters object
        """
        new = []
        for p in self:
            if p.name in fix:
                newparam = p.replace(fix=fix[p.name])
            else:
                newparam = p
            new.append(newparam)
        return Parameters(tuple(new))

    @property
    def fixed(self) -> Parameters:
        """All fixed parameters"""
        fixed = [p for p in self._params if p.fix]
        return Parameters(tuple(fixed))

    @property
    def nonfixed(self) -> Parameters:
        """All non-fixed parameters"""
        nonfixed = [p for p in self._params if not p.fix]
        return Parameters(tuple(nonfixed))

    def __add__(self, other: Union[Parameter, Parameters, Sequence[Parameter]]) -> Parameters:
        if isinstance(other, Parameter):
            return Parameters.create(self._params + (other,))
        elif isinstance(other, Parameters):
            return Parameters.create(self._params + other._params)
        elif isinstance(other, Sequence):
            return Parameters.create(self._params + tuple(other))
        else:
            raise ValueError(f"Cannot add {other} to Parameters")

    def __radd__(self, other: Union[Parameter, Sequence[Parameter]]) -> Parameters:
        if isinstance(other, Parameter):
            return Parameters.create((other,) + self._params)
        elif isinstance(other, Sequence):
            return Parameters.create(tuple(other) + self._params)
        else:
            raise ValueError(f"Cannot add {other} to Parameters")

    def __eq__(self, other):
        if hash(self) != hash(other):
            return False
        if not isinstance(other, Parameters):
            return False
        if len(self) != len(other):
            return False
        for p1, p2 in zip(self, other):
            if p1 != p2:
                return False
        return True

    @cache_method
    def __hash__(self):
        return hash(self._params)

    def to_dict(self):
        params = tuple(p.to_dict() for p in self._params)
        return {'parameters': params}

    @classmethod
    def from_dict(cls, d):
        params = tuple(Parameter.from_dict(p) for p in d['parameters'])
        return cls(parameters=params)

    def __repr__(self):
        if len(self) == 0:
            return "Parameters()"
        return (
            self.to_dataframe().replace(float("inf"), "∞").replace(-float("inf"), "-∞").to_string()
        )

    def _repr_html_(self):
        if len(self) == 0:
            return "Parameters()"
        else:
            return (
                self.to_dataframe()
                .replace(float("inf"), "∞")
                .replace(-float("inf"), "-∞")
                .to_html()
            )



import re
from inspect import getdoc
from typing import List

def create_r_doc(func):
    doc = getdoc(func)

    if not doc:
        return f'#\' @title\n' \
               f'#\' {func.__name__}\n' \
               f'#\' \n' \
               f'#\' @export'

    doc_dict = split_doc_to_subtypes(doc)

    doc_str = f'@title\n{func.__name__}\n\n@description\n'

    for row in doc_dict['description']:
        is_list = row.startswith('-')
        if is_list:
            row = re.sub('^- ', '* ', row)
        if row.startswith('.. '):
            row = re.sub(r'^\.\. ', '', row)
        doc_str += f'{py_to_r_str(row)}\n'

    if 'params' in doc_dict.keys():
        doc_str += create_r_params(doc_dict['params'])
    if 'returns' in doc_dict.keys():
        doc_str += create_r_returns(doc_dict['returns']) + '\n\n'
    if 'examples' in doc_dict.keys():
        doc_str += create_r_example(doc_dict['examples']) + '\n'
    if 'notes' in doc_dict.keys():
        doc_str += '@note\n'
        doc_str += ''.join(doc_dict['notes']) + '\n\n'
    if 'see_also' in doc_dict.keys():
        doc_str += '@seealso\n'
        doc_str += '\n\n'.join(doc_dict['see_also']) + '\n\n'

    r_doc = ''
    for row in doc_str.split('\n'):
        r_doc += f'#\' {row}\n'

    r_doc += f'#\' @export'
    return r_doc


def create_r_params(doc_list):
    doc_str = ''
    skip = False
    for i, row in enumerate(doc_list):
        if skip:
            skip = False
            continue
        type_declare_pattern = re.compile(r'([\w0-9]+) : ([\w0-9]+)')
        if type_declare_pattern.match(row):
            type_declare = row.split(' : ')
            doc_str += f'@param {type_declare[0]} ({py_to_r_str(type_declare[1])})'
        elif row == 'args' or row == 'kwargs':
            doc_str += f'@param ... {doc_list[i+1]}\n'
            skip = True
        else:
            doc_str += f' {py_to_r_str(row)}\n'
    return doc_str


def create_r_returns(doc_list):
    doc_list_trim = list(filter(None, doc_list))
    type_py = doc_list_trim[0]
    type_r = py_to_r_str(type_py)

    doc_str = f'@return ({type_r}) '
    doc_str += ' '.join(py_to_r_str(row) for row in doc_list_trim[1:])

    return doc_str

pattern_start = re.compile(r'>>> |^\.\.\. ')
pattern_methods = re.compile(r'([A-Za-z_]\d*)\.([A-Za-z_]\d*)(?=(?:[^"]|"[^"]*")*$)(?=(?:[^\']|\'[^\']*\')*$)')
pattern_list_idx = re.compile(r'\w\[(\d+)]')
pattern_list = re.compile(r'\[([\'\"\w(),\s]+)](?=(?:[^"]|"[^"]*")*$)(?=(?:[^\']|\'[^\']*\')*$)')
pattern_dict = re.compile(r'{(([\'\"]*[\w\d()]+[\'\"]*: [\'\"]*[\w\d]+\'*,*\s*)+)}')
pattern_doctest = re.compile(r'\s+# doctest:.*')

def line_transpile_py_to_r(line: str) -> str:
    line = py_to_r_str(line, example=True)
    line = re.sub(' = ', ' <- ', line)

    # Substitute any . to $, e.g. model.parameters -> model$parameters
    line = re.sub(pattern_methods, r'\1$\2', line)

    # Substitute """ or ''' to " for multiline strings
    line = re.sub(r'"""', r'"', line)
    line = re.sub(r"'''", r'"', line)

    # Check if line has list subscript
    if re.search(r'\w\[', line):
        idx = re.search(pattern_list_idx, line)
        if idx:
            # Increase idx by 1, e.g. list[0] -> list[1]
            line = re.sub(idx.group(1), f'{int(idx.group(1)) + 1}', line)
    else:
        # Substitute any [] to c(), e.g. ['THETA(1)'] -> c('THETA(1)')
        line = re.sub(pattern_list, r'c(\1)', line)

    # Check if line contains python dict
    dict_py = re.search(pattern_dict, line)
    if dict_py:
        # Substitute {} to list(), e.g. {'EONLY': 1} -> list('EONLY'=1)
        dict_r = f'list({dict_py.group(1).replace(": ", "=", )})'
        line = line.replace('{' + f'{dict_py.group(1)}' + '}', dict_r)

    # Replace len() with length()
    line = line.replace(r'len(', 'length(')

    # Remove doctest comments
    line = re.sub(pattern_doctest, '', line)
    return line

def is_import_line(line: str):
    return line.startswith('import ') or ' import ' in line

def transpile_py_to_r(lines: List[str]):
    # must_load_library = any(map(is_import_line, lines))
    filtered_lines = [line for line in lines if not(is_import_line(line))]

    transpiled_lines = map(
        lambda line: line_transpile_py_to_r(re.sub(pattern_start, '', line)),
        filtered_lines,
    )

    return transpiled_lines
    # return ['library(pharmr)', *transpiled_lines] if must_load_library else list(transpiled_lines)

def create_r_example(doc_list):
    # Check for rows that starts with ... or >>>
    doc_code = [row for row in doc_list if row.startswith('>>>') or re.match(r'^\.\.\.\s+[$\w\d]', row)]

    doc_code_r = map(
        lambda line: line  + '\n',
        transpile_py_to_r([
            re.sub(pattern_start, '', line) for line in doc_code
        ])
    )

    return '@examples\n\\dontrun{\n' + ''.join(doc_code_r) + '}'


def split_doc_to_subtypes(doc_str):
    doc_split = doc_str.split('\n')

    doc_titles = {'Parameters': 'params',
                  'Returns': 'returns',
                  'Return': 'returns',
                  'Results': 'returns',
                  'Example': 'examples',
                  'Examples': 'examples',
                  'Notes': 'notes',
                  'See also': 'see_also',
                  'See Also': 'see_also'}

    doc_type_current = 'description'

    doc_dict = {}
    doc_dict[doc_type_current] = []

    for row in doc_split:
        if row in doc_titles.keys() and doc_titles[row] not in doc_dict.keys():
            doc_type_current = doc_titles[row]
            doc_dict[doc_type_current] = []
            continue

        if not row.startswith('--'):
            doc_dict[doc_type_current].append(row.strip())

    return doc_dict


def py_to_r_arg(arg):
    py_to_r_dict = {'None': 'NULL',
                    'True': 'TRUE',
                    'False': 'FALSE',
                    '': '\'\''}

    try:
        return py_to_r_dict[str(arg)]
    except KeyError:
        if isinstance(arg, str):
            return f'\'{arg}\''
        else:
            return arg


def py_to_r_str(arg, example=False):
    args = {'None': 'NULL',
            'True': 'TRUE',
            'False': 'FALSE'}

    types = {r'\bint\b': 'integer',
             'float': 'numeric',
             r'\bbool\b': 'logical',
             r'\blist\b': 'vector',
             r'\bdict\b': 'list',
             'dictionary': 'list',
             'pd.DataFrame': 'data.frame',
             'pd.Series': 'data.frame',
             r'List\[Model\]': 'vector of Models'}  # FIXME: more general pattern

    latex = {r'\\mathsf': '',
             r'\\cdot': '*',
             r'\\text': '',
             r'\\frac': 'frac',
             r'\\log': 'log',
             r'\\exp': 'exp',
             r'\\min': 'min',
             r'\\max': 'max',
             r'\\epsilon': 'epsilon'}

    py_to_r_dict = {**args, **types, **latex}

    if not example:
        py_to_r_dict = {**py_to_r_dict, r'\[([0-9]+)\]_*': r'(\1)'}

    arg_sub = arg
    for key, value in py_to_r_dict.items():
        arg_sub = re.sub(key, value, arg_sub)

    return arg_sub




import shutil
from pathlib import Path

import pytest

from pharmpy.internals.fs.cwd import chdir
from pharmpy.tools import run_amd


def _model_count(rundir: Path):
    return sum(
        map(
            lambda path: 0 if path.name in ['.lock', '.datasets'] else 1,
            ((rundir / 'models').iterdir()),
        )
    )


@pytest.mark.parametrize(
    'strategy, subrundir',
    [
        (
            'all',
            [
                'modelfit',
                'modelsearch',
                'iivsearch',
                'ruvsearch',
                'iovsearch',
                'allometry',
                'covsearch_exploratory',
            ],
        ),
        (
            'retries',
            [
                'modelfit',
                'modelsearch',
                'iivsearch',
                'ruvsearch',
                'iovsearch',
                'allometry',
                'covsearch_exploratory',
                'rerun_iivsearch',
                'rerun_ruvsearch',
            ],
        ),
    ],
)
@pytest.mark.filterwarnings(
    'ignore:.*Adjusting initial estimates to create positive semidefinite omega/sigma matrices.',
    'ignore::UserWarning',
)
def test_amd(tmp_path, testdata, strategy, subrundir):
    with chdir(tmp_path):
        shutil.copy2(testdata / 'nonmem' / 'models' / 'moxo_simulated_amd.csv', '.')
        shutil.copy2(testdata / 'nonmem' / 'models' / 'moxo_simulated_amd.datainfo', '.')
        input = 'moxo_simulated_amd.csv'
        res = run_amd(
            input,
            modeltype='basic_pk',
            administration='oral',
            search_space='PERIPHERALS(1)',
            strategy=strategy,
            occasion='VISI',
            strictness='minimization_successful or rounding_errors',
            retries_strategy='skip',
        )

        rundir = tmp_path / 'amd_dir1'
        assert rundir.is_dir()
        assert (rundir / 'results.json').exists()
        assert (rundir / 'results.csv').exists()

        for dir in subrundir:
            dir = rundir / dir
            assert _model_count(dir) >= 1

        assert len(res.summary_tool) == len(subrundir)
        assert len(res.summary_models) >= 1
        assert len(res.summary_individuals_count) >= 1


# def test_structure_mechanistic_exploratory(tmp_path, testdata):
#     with chdir(tmp_path):
#         shutil.copy2(testdata / 'nonmem' / 'models' / 'moxo_simulated_amd.csv', '.')
#         shutil.copy2(testdata / 'nonmem' / 'models' / 'moxo_simulated_amd.datainfo', '.')
#         input = 'moxo_simulated_amd.csv'
#         res = run_amd(
#             input,
#             modeltype='basic_pk',
#             administration='oral',
#             search_space='PERIPHERALS(1);COVARIATE(CL,WT,pow);COVARIATE?(VC,AGE,exp);COVARIATE?(Q,SEX,cat)',
#             mechanistic_covariates=["AGE"],
#             occasion='VISI',
#             strictness='minimization_successful or rounding_errors',
#             retries_strategy='skip',
#         )

#         rundir = tmp_path / 'amd_dir1'
#         assert rundir.is_dir()
#         assert (rundir / 'results.json').exists()
#         assert (rundir / 'results.csv').exists()
#         subrundir = [
#             'modelfit',
#             'modelsearch',
#             'iivsearch',
#             'ruvsearch',
#             'iovsearch',
#             'allometry',
#             'covsearch_structural',
#             'covsearch_mechanistic',
#             'covsearch_exploratory',
#         ]
#         for dir in subrundir:
#             dir = rundir / dir
#             assert _model_count(dir) >= 1

#         assert len(res.summary_tool) == len(subrundir) - 1  # Mechanistic/Exploratory grouped as one
#         assert len(res.summary_models) >= 1
#         assert len(res.summary_individuals_count) >= 1



from pathlib import Path

import pytest

from pharmpy.deps import sympy
from pharmpy.model import Assignment, Model
from pharmpy.model.external.nonmem.dataset import read_nonmem_dataset
from pharmpy.modeling import convert_model, create_basic_pk_model, create_symbol, load_example_model

tabpath = Path(__file__).resolve().parent.parent / 'testdata' / 'nonmem' / 'pheno_real_linbase.tab'
lincorrect = read_nonmem_dataset(
    tabpath,
    ignore_character='@',
    colnames=['ID', 'G11', 'G21', 'H11', 'CIPREDI', 'DV', 'PRED', 'RES', 'WRES'],
)


@pytest.mark.parametrize(
    'stem,force_numbering,symbol_name', [('DV', False, 'DV1'), ('X', False, 'X'), ('X', True, 'X1')]
)
def test_create_symbol(load_model_for_test, testdata, stem, force_numbering, symbol_name):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno_real.mod')
    symbol = create_symbol(model, stem=stem, force_numbering=force_numbering)

    assert symbol.name == symbol_name


def test_to_generic_model(load_model_for_test, testdata):
    path = testdata / 'nonmem' / 'pheno.mod'
    nm_model = load_model_for_test(path)
    model = convert_model(nm_model, 'generic')

    assert model.parameters == nm_model.parameters
    assert model.random_variables == nm_model.random_variables
    assert model.name == nm_model.name
    assert model.statements == nm_model.statements
    assert isinstance(model, Model)


def test_model_equality():
    pheno1 = load_example_model("pheno")
    assert pheno1 == pheno1

    pheno2 = load_example_model("pheno")
    assert pheno2 == pheno2

    assert pheno1 == pheno2

    pheno_linear1 = load_example_model("pheno_linear")
    assert pheno_linear1 == pheno_linear1

    pheno_linear2 = load_example_model("pheno_linear")
    assert pheno_linear2 == pheno_linear2

    assert pheno_linear1 == pheno_linear2

    assert pheno1 != pheno_linear1
    assert pheno1 != pheno_linear2
    assert pheno2 != pheno_linear1
    assert pheno2 != pheno_linear2

    assert hash(pheno1) != hash(pheno_linear1)


def test_replace(load_model_for_test, testdata):
    path = testdata / 'nonmem' / 'pheno.mod'
    model = load_model_for_test(path)
    sset = model.statements
    cl = sset.find_assignment('CL')
    sset_new = sset.reassign(cl.symbol, cl.expression + sympy.Symbol('x'))
    with pytest.raises(ValueError, match='Symbol x is not defined'):
        model.replace(statements=sset_new)

    x_assignment = Assignment(sympy.Symbol('x'), sympy.Float(1))
    model.replace(statements=x_assignment + sset_new)

    sset_new = sset_new.before_odes + x_assignment + sset_new.ode_system + sset_new.after_odes
    with pytest.raises(ValueError, match='Symbol x defined after being used'):
        model.replace(statements=sset_new)

    sset_new = sset.reassign(cl.symbol, cl.expression + sympy.Symbol('TIME'))
    model.replace(statements=sset_new)


def test_dict(load_model_for_test, testdata):
    path = testdata / 'nonmem' / 'pheno.mod'
    model = load_model_for_test(path)
    d = model.to_dict()
    model2 = Model.from_dict(d)
    assert model == model2


def test_dict_generic(load_model_for_test, testdata):
    path = testdata / 'nonmem' / 'pheno.dta'
    model = create_basic_pk_model('iv', dataset_path=path)
    assert model.model_code



from __future__ import annotations

import re
from dataclasses import dataclass
from typing import Any, Optional

from pharmpy.deps import pandas as pd
from pharmpy.tools import read_modelfit_results
from pharmpy.workflows import Results


@dataclass(frozen=True)
class CrossvalResults(Results):
    """Crossval results class"""

    runs: Optional[Any] = None
    prediction_ofv_sum: Optional[Any] = None


def calculate_results(estimation_results, prediction_results):
    """Calculate crossval results"""
    est_ofvs = [res.ofv for res in estimation_results]
    pred_ofvs = [res.ofv for res in prediction_results]
    runs = pd.DataFrame(
        {'estimation_ofv': est_ofvs, 'prediction_ofv': pred_ofvs},
        index=pd.RangeIndex(start=1, stop=len(est_ofvs) + 1),
    )

    return CrossvalResults(runs=runs, prediction_ofv_sum=sum(pred_ofvs))


def psn_crossval_results(path):
    def atoi(text):
        return int(text) if text.isdigit() else text

    def natural_keys(text):
        return [atoi(c) for c in re.split(r'(\d+)', text)]

    est_paths = [str(p) for p in (path / 'm1').glob('est_model*.mod')]
    est_paths.sort(key=natural_keys)
    pred_paths = [str(p) for p in (path / 'm1').glob('pred_model*.mod')]
    pred_paths.sort(key=natural_keys)
    ests = [read_modelfit_results(path) for path in est_paths]
    preds = [read_modelfit_results(path) for path in pred_paths]

    res = calculate_results(ests, preds)
    return res



from __future__ import annotations

from typing import TYPE_CHECKING, List, Optional, Union

from pharmpy.internals.expr.parse import parse as parse_expr
from pharmpy.model import Assignment, Model, Parameter, Parameters

from .expressions import _create_symbol
from .odes import find_clearance_parameters, find_volume_parameters

if TYPE_CHECKING:
    import sympy
else:
    from pharmpy.deps import sympy


def add_allometry(
    model: Model,
    allometric_variable: Optional[Union[str, sympy.Expr]] = None,
    reference_value: Union[str, int, float, sympy.Expr] = 70,
    parameters: Optional[List[Union[str, sympy.Expr]]] = None,
    initials: Optional[List[Union[int, float]]] = None,
    lower_bounds: Optional[List[Union[int, float]]] = None,
    upper_bounds: Optional[List[Union[int, float]]] = None,
    fixed: bool = True,
):
    """Add allometric scaling of parameters

    Add an allometric function to each listed parameter. The function will be
    P=P*(X/Z)**T where P is the parameter, X the allometric_variable, Z the reference_value
    and T is a theta. Default is to automatically use clearance and volume parameters.

    If there already exists a covariate effect (or allometric scaling) on a parameter
    with the specified allometric variable, nothing will be added.

    If no allometric variable is specified, it will be extracted from the dataset based on
    the descriptor "body weight".

    Parameters
    ----------
    model : Model
        Pharmpy model
    allometric_variable : str or sympy.Expr
        Value to use for allometry (X above)
    reference_value : str, int, float or sympy.Expr
        Reference value (Z above)
    parameters : list
        Parameters to use or None (default) for all available CL, Q and V parameters
    initials : list
        Initial estimates for the exponents. Default is to use 0.75 for CL and Qs and 1 for Vs
    lower_bounds : list
        Lower bounds for the exponents. Default is 0 for all parameters
    upper_bounds : list
        Upper bounds for the exponents. Default is 2 for all parameters
    fixed : bool
        Whether the exponents should be fixed

    Return
    ------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, add_allometry, remove_covariate_effect
    >>> model = load_example_model("pheno")
    >>> model = remove_covariate_effect(model, 'CL', 'WGT')
    >>> model = remove_covariate_effect(model, 'V', 'WGT')
    >>> model = add_allometry(model, allometric_variable='WGT')
    >>> model.statements.before_odes
            ⎧TIME  for AMT > 0
            ⎨
    BTIME = ⎩ 0     otherwise
    TAD = -BTIME + TIME
    TVCL = PTVCL
    TVV = PTVV
          ⎧TVV⋅(THETA₃ + 1)  for APGR < 5
          ⎨
    TVV = ⎩      TVV           otherwise
               ETA₁
    CL = TVCL⋅ℯ
                 ALLO_CL
            ⎛WGT⎞
         CL⋅⎜───⎟
    CL =    ⎝ 70⎠
             ETA₂
    V = TVV⋅ℯ
               ALLO_V
          ⎛WGT⎞
        V⋅⎜───⎟
    V =   ⎝ 70⎠
    S₁ = V

    """
    from pharmpy.modeling import has_covariate_effect

    if allometric_variable is None:
        try:
            allometric_variable = model.datainfo.descriptorix["body weight"][0].name
        except IndexError:
            raise ValueError(
                "No allometric variable could be found. Try setting the allometric_variable argument"
            )

    variable = parse_expr(allometric_variable)
    reference = parse_expr(reference_value)

    parsed_parameters = []

    if parameters is not None:
        parsed_parameters = [parse_expr(p) for p in parameters]

    if parameters is None or initials is None:
        cls = find_clearance_parameters(model)
        vcs = find_volume_parameters(model)

        if parameters is None:
            parsed_parameters = cls + vcs

        if initials is None:
            # Need to understand which parameter is CL or Q and which is V
            initials = []
            for p in parsed_parameters:
                if p in cls:
                    initials.append(0.75)
                elif p in vcs:
                    initials.append(1.0)

    if not parsed_parameters:
        raise ValueError("No parameters provided")

    if lower_bounds is None:
        lower_bounds = [0.0] * len(parsed_parameters)
    if upper_bounds is None:
        upper_bounds = [2.0] * len(parsed_parameters)

    if not (len(parsed_parameters) == len(initials) == len(lower_bounds) == len(upper_bounds)):
        raise ValueError("The number of parameters, initials and bounds must be the same")
    sset = model.statements
    params = list(model.parameters)
    for p, init, lower, upper in zip(parsed_parameters, initials, lower_bounds, upper_bounds):
        if not has_covariate_effect(model, str(p), str(variable)):
            symb = _create_symbol(
                sset, params, model.random_variables, model.datainfo, f'ALLO_{p.name}', False
            )
            param = Parameter(symb.name, init=init, lower=lower, upper=upper, fix=fixed)
            params.append(param)
            expr = p * (variable / reference) ** param.symbol
            new_ass = Assignment(p, expr)
            ind = sset.find_assignment_index(p)
            sset = sset[0 : ind + 1] + new_ass + sset[ind + 1 :]
    parameters = Parameters.create(params)
    model = model.replace(statements=sset, parameters=parameters)
    model = model.update_source()

    return model



from pharmpy.model import Bolus, Infusion
from pharmpy.modeling import create_basic_pk_model


def test_create_start_model(testdata):
    path = testdata / 'nonmem' / 'pheno.dta'
    model = create_basic_pk_model(dataset_path=path, administration='iv')
    sep = model.datainfo.separator
    assert sep == "\\s+"
    assert len(model.dataset.columns) == 8
    assert len(model.parameters) == 6
    assert 'POP_CL' in model.parameters
    assert 'POP_MAT' not in model.parameters
    assert model.statements.ode_system.dosing_compartments[0].doses[0] == Bolus.create("AMT")
    model = create_basic_pk_model(dataset_path=path, administration='oral')
    assert 'IIV_MAT' in model.parameters
    assert 'POP_CL' in model.parameters
    assert 'POP_MAT' in model.parameters

    path_2 = testdata / 'nonmem' / 'modeling' / 'pheno_zero_order.csv'
    model = create_basic_pk_model(dataset_path=path_2, administration='iv')
    assert model.statements.ode_system.dosing_compartments[0].doses[0] == Infusion.create(
        "AMT", duration="D1"
    )



import shutil
from dataclasses import replace

import pytest

from pharmpy.internals.fs.cwd import chdir
from pharmpy.modeling import remove_parameter_uncertainty_step, transform_blq
from pharmpy.tools import read_modelfit_results
from pharmpy.tools.ruvsearch.results import psn_resmod_results
from pharmpy.tools.ruvsearch.tool import _create_dataset, create_workflow, validate_input
from pharmpy.workflows import ModelEntry, Workflow


def test_filter_dataset(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem/pheno_pd.mod')
    res = read_modelfit_results(testdata / 'nonmem/pheno_pd.mod')
    model_entry = ModelEntry.create(model, modelfit_results=res)
    cwres = _create_dataset(model_entry, dv=2)
    expected_cwres = [-1.15490, 0.95703, -0.85365, 0.42327]
    assert cwres['DV'].tolist() == expected_cwres


def test_resmod_results(testdata):
    res = psn_resmod_results(testdata / 'psn' / 'resmod_dir1')
    assert list(res.cwres_models['dOFV']) == [
        -1.31,
        -3.34,
        -13.91,
        -18.54,
        -8.03,
        -4.20,
        -0.25,
        -1.17,
        -0.00,
        -0.09,
        -2.53,
        -3.12,
        -3.60,
        -25.62,
        -7.66,
        -0.03,
        -5.53,
    ]


def test_resmod_results_dvid(testdata):
    res = psn_resmod_results(testdata / 'psn' / 'resmod_dir2')
    df = res.cwres_models
    assert df['dOFV'].loc[1, '1', 'autocorrelation'] == -0.74
    assert df['dOFV'].loc[1, 'sum', 'tdist'] == -35.98


def test_create_workflow():
    assert isinstance(create_workflow(), Workflow)


def test_create_workflow_with_model(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'ruvsearch' / 'mox3.mod')
    remove_parameter_uncertainty_step(model)
    assert isinstance(create_workflow(model=model), Workflow)


def test_validate_input():
    validate_input()


def test_validate_input_with_model(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'ruvsearch' / 'mox3.mod')
    res = read_modelfit_results(testdata / 'nonmem' / 'ruvsearch' / 'mox3.mod')
    model = remove_parameter_uncertainty_step(model)
    validate_input(model=model, results=res)


def test_create_dataset(load_model_for_test, testdata, tmp_path):
    model = load_model_for_test(testdata / 'nonmem' / 'ruvsearch' / 'mox3.mod')
    res = read_modelfit_results(testdata / 'nonmem' / 'ruvsearch' / 'mox3.mod')
    model_entry = ModelEntry.create(model, modelfit_results=res)
    df = _create_dataset(model_entry, dv=None)

    assert len(df) == 1006
    assert (df['DV'] != 0).all()

    with chdir(tmp_path):
        for path in (testdata / 'nonmem' / 'ruvsearch').glob('mox3.*'):
            shutil.copy2(path, tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'ruvsearch' / 'moxo_simulated_resmod.csv', tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'ruvsearch' / 'mytab', tmp_path)

        # Introduce 0 in CWRES to mimic rows BLQ
        with open('mytab') as f:
            mytab_new = f.read().replace('-2.4366E+00', '0.0000E+00')

        with open('mytab', 'w') as f:
            f.write(mytab_new)

        model = load_model_for_test('mox3.mod')
        res = read_modelfit_results('mox3.mod')

        model = transform_blq(model, method='m3', lloq=0.05)
        model_entry = ModelEntry.create(model, modelfit_results=res)

        df = _create_dataset(model_entry, dv=None)

        assert len(df) == 1005
        assert (df['DV'] != 0).all()


@pytest.mark.parametrize(
    ('model_path', 'arguments', 'exception', 'match'),
    [
        (
            None,
            dict(groups=3.1415),
            TypeError,
            'Invalid `groups`',
        ),
        (
            None,
            dict(groups=0),
            ValueError,
            'Invalid `groups`',
        ),
        (
            None,
            dict(p_value='x'),
            TypeError,
            'Invalid `p_value`',
        ),
        (
            None,
            dict(p_value=1.01),
            ValueError,
            'Invalid `p_value`',
        ),
        (
            None,
            dict(skip='ABC'),
            TypeError,
            'Invalid `skip`',
        ),
        (
            None,
            dict(skip=1),
            TypeError,
            'Invalid `skip`',
        ),
        (
            None,
            dict(skip=['IIV_on_RUV', 'power', 'time', 0]),
            TypeError,
            'Invalid `skip`',
        ),
        (
            None,
            dict(skip=['IIV_on_RUV', 'power', 'time']),
            TypeError,
            'Invalid `skip`',
        ),
        (
            None,
            dict(model=1),
            TypeError,
            'Invalid `model`',
        ),
    ],
)
def test_validate_input_raises(
    load_model_for_test,
    testdata,
    model_path,
    arguments,
    exception,
    match,
):
    model = load_model_for_test(testdata.joinpath(*model_path)) if model_path else None

    kwargs = {'model': model, **arguments}

    with pytest.raises(exception, match=match):
        validate_input(**kwargs)

    validate_input(None, skip=['IIV_on_RUV', 'power'])


def test_validate_input_raises_modelfit_results(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno.mod')

    with pytest.raises(ValueError, match="modelfit results must be provided"):
        validate_input(model=model, results=None)


def test_validate_input_raises_cwres(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'ruvsearch' / 'mox3.mod')
    res = read_modelfit_results(testdata / 'nonmem' / 'ruvsearch' / 'mox3.mod')
    model = remove_parameter_uncertainty_step(model)
    modelfit_results = replace(res, residuals=res.residuals.drop(columns=['CWRES']))

    with pytest.raises(ValueError, match="CWRES"):
        validate_input(model=model, results=modelfit_results)


def test_validate_input_raises_cipredi(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'ruvsearch' / 'mox3.mod')
    res = read_modelfit_results(testdata / 'nonmem' / 'ruvsearch' / 'mox3.mod')
    model = remove_parameter_uncertainty_step(model)
    modelfit_results = replace(res, predictions=res.predictions.drop(columns=['CIPREDI']))

    with pytest.raises(ValueError, match="IPRED"):
        validate_input(model=model, results=modelfit_results)


def test_validate_input_raises_ipred(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno_real.mod')
    res = read_modelfit_results(testdata / 'nonmem' / 'pheno_real.mod')
    model = remove_parameter_uncertainty_step(model)
    modelfit_results = replace(res, predictions=res.predictions.drop(columns=['IPRED']))

    with pytest.raises(ValueError, match="IPRED"):
        validate_input(model=model, results=modelfit_results)



import os as _os
import shutil as _shutil
import sys
import warnings as _warnings
import weakref as _weakref
from tempfile import mkdtemp


# NOTE: This is copied from the Python 3.10 implementation to expose the
# ignore_cleanup_errors flag and patch PermissionError handling for Windows
# lock files. Here is the original source link:
#
# https://github.com/python/cpython/blob/b494f5935c92951e75597bfe1c8b1f3112fec270/Lib/tempfile.py#L778-L848
#
# The whole thing can be replaced by `from tempfile import TemporaryDirectory`
# once Python fixes their implementation
class TemporaryDirectory:
    """Create and return a temporary directory.  This has the same
    behavior as mkdtemp but can be used as a context manager.  For
    example:

        with TemporaryDirectory() as tmpdir:
            ...

    Upon exiting the context, the directory and everything contained
    in it are removed.
    """

    def __init__(self, suffix=None, prefix=None, dir=None, ignore_cleanup_errors=False):
        self.name = mkdtemp(suffix, prefix, dir)
        self._ignore_cleanup_errors = ignore_cleanup_errors
        self._finalizer = _weakref.finalize(
            self,
            self._cleanup,
            self.name,
            warn_message="Implicitly cleaning up {!r}".format(self),
            ignore_errors=self._ignore_cleanup_errors,
        )

    @classmethod
    def _rmtree(cls, name, ignore_errors=False):
        def onerror(func, path, exc_info):
            if sys.version_info.minor >= 12:
                exc = exc_info
            else:
                # This was deprecated in Python 3.12
                exc = exc_info[0]
            if issubclass(exc, PermissionError):

                def resetperms(path):
                    try:
                        _os.chflags(path, 0)  # pyright: ignore [reportGeneralTypeIssues]
                    except AttributeError:
                        pass
                    _os.chmod(path, 0o700)

                try:
                    if path != name:
                        resetperms(_os.path.dirname(path))
                    resetperms(path)

                    try:
                        _os.unlink(path)
                    except IsADirectoryError:
                        cls._rmtree(path, ignore_errors=ignore_errors)
                    except PermissionError as pe:
                        # PermissionError is raised on FreeBSD for directories
                        # and by Windows on lock files used by other processes
                        try:
                            cls._rmtree(path, ignore_errors=ignore_errors)
                        except NotADirectoryError:
                            # NOTE: This is raised if PermissionError did not
                            # correspond to a IsADirectoryError, e.g. on
                            # Windows.
                            if not ignore_errors:
                                raise pe
                except FileNotFoundError:
                    pass
            elif issubclass(exc, FileNotFoundError):
                pass
            else:
                if not ignore_errors:
                    raise

        if sys.version_info.minor >= 12:
            _shutil.rmtree(name, onexc=onerror)  # pyright: ignore
        else:
            # This was deprecated in Python 3.12
            _shutil.rmtree(name, onerror=onerror)

    @classmethod
    def _cleanup(cls, name, warn_message, ignore_errors=False):
        cls._rmtree(name, ignore_errors=ignore_errors)
        _warnings.warn(warn_message, ResourceWarning)

    def __repr__(self):
        return "<{} {!r}>".format(self.__class__.__name__, self.name)

    def __enter__(self):
        return self.name

    def __exit__(self, exc, value, tb):
        self.cleanup()

    def cleanup(self):
        if self._finalizer.detach() or _os.path.exists(self.name):
            self._rmtree(self.name, ignore_errors=self._ignore_cleanup_errors)



import re

import pytest

from pharmpy.workflows import Task, Workflow, WorkflowBuilder


@pytest.fixture
def tasks():
    t1 = Task('t1', 'func', 'input')
    t2 = Task('t2', 'func', t1)
    t3 = Task('t3', 'func', t1)
    t4 = Task('t4', 'func', [t2, t3])
    return t1, t2, t3, t4


def test_add_tasks(tasks):
    wb = WorkflowBuilder()

    t1, t2, t3, _ = tasks

    wb.add_task(t1)
    assert len(wb) == 1

    wb.add_task(t2)
    wb.add_task(t3)
    assert len(wb) == 3


def test_add_tasks_workflow(tasks):
    t1, t2, t3, _ = tasks

    wb_sequential = WorkflowBuilder(tasks=[t1])
    wb_sequential.add_task(t2, predecessors=t1)
    wb_sequential.add_task(t3, predecessors=t1)
    assert list(Workflow(wb_sequential).get_successors(t1)) == [t2, t3]

    t4 = Task('t4', 'func', 'input')
    wb_t4 = WorkflowBuilder(tasks=[t4])

    wb_sequential.insert_workflow(wb_t4)
    assert list(Workflow(wb_sequential).get_predecessors(t4)) == [t2, t3]
    assert len(wb_sequential) == 4

    wb_parallel = WorkflowBuilder(tasks=[t1])
    wb_parallel.add_task(t2, predecessors=t1)
    wb_parallel.add_task(t3, predecessors=t1)
    wb_parallel.insert_workflow(Workflow(wb_t4), predecessors=[])
    assert not list(Workflow(wb_parallel).get_predecessors(t4))
    assert len(wb_parallel) == 4


def test_get_output(tasks):
    t1, t2, t3, t4 = tasks
    wb = WorkflowBuilder(tasks=[t1])
    wb.add_task(t2, predecessors=t1)
    wb.add_task(t3, predecessors=t1)

    assert Workflow(wb).output_tasks == [t2, t3]

    wb.add_task(t4, predecessors=[t2, t3])

    assert Workflow(wb).output_tasks == [t4]


def test_as_dask_dict(tasks):
    wb = WorkflowBuilder()

    t1, t2, t3, t4 = tasks

    wb.add_task(t1)
    wb.add_task(t2, predecessors=t1)
    wb.add_task(t3, predecessors=t1)
    wb.add_task(t4, predecessors=[t2, t3])

    wf_dict = Workflow(wb).as_dask_dict()
    assert re.match(
        r"\{'.*': \('func', 'input'\), '.*': \('func', t1, '.*'\), '.*': \('func', t1, '.*'\), 'results': \('func', \[t2, t3\], '.*', '.*'\)\}",  # noqa
        str(wf_dict),
    )


def test_as_dask_dict_raises(tasks):
    wb = WorkflowBuilder(tasks=tasks)

    with pytest.raises(ValueError, match='.*one output task.*'):
        Workflow(wb).as_dask_dict()


def test_insert_workflow(tasks):
    t1, t2, t3, t4 = tasks

    wb = WorkflowBuilder(tasks=[t1])
    wb.insert_workflow(WorkflowBuilder(tasks=[t2, t3]))
    wb.insert_workflow(WorkflowBuilder(tasks=[t4]))

    assert set(wb.input_tasks) == {t1}
    assert set(wb.output_tasks) == {t4}


def test_insert_workflow_with_single_predecessors(tasks):
    t1, t2, t3, t4 = tasks

    wb = WorkflowBuilder(tasks=[t1])
    wb.insert_workflow(WorkflowBuilder(tasks=[t2, t3]), predecessors=t1)
    wb.insert_workflow(WorkflowBuilder(tasks=[t4]), predecessors=t2)

    assert set(wb.input_tasks) == {t1}
    assert set(wb.output_tasks) == {t3, t4}


def test_insert_workflow_with_predecessors(tasks):
    t1, t2, t3, t4 = tasks

    wb = WorkflowBuilder(tasks=[t1])
    wb.insert_workflow(WorkflowBuilder(tasks=[t2, t3]), predecessors=[t1])
    wb.insert_workflow(WorkflowBuilder(tasks=[t4]), predecessors=[t2, t3])

    assert set(wb.input_tasks) == {t1}
    assert set(wb.output_tasks) == {t4}


def test_insert_workflow_nm():
    t1, t2, t3, t4, t5 = map(lambda i: Task(f't{i}', lambda: 0), range(5))
    wb1 = WorkflowBuilder(tasks=[t1, t2])
    wb2 = WorkflowBuilder(tasks=[t3, t4, t5])
    with pytest.raises(ValueError, match='.*not supported.*'):
        wb1.insert_workflow(wb2)


def test_get_upstream_tasks():
    t1, t2, t3, t4, t5 = map(lambda i: Task(f't{i}', lambda: 0), range(5))
    wb = WorkflowBuilder(tasks=[t1, t2])
    wb.insert_workflow(WorkflowBuilder(tasks=[t3]))
    wb.insert_workflow(WorkflowBuilder(tasks=[t4, t5]))

    wf = Workflow(wb)
    assert set(wf.get_upstream_tasks(t4)) == set(wf.get_upstream_tasks(t5)) == {t1, t2, t3}
    assert set(wf.get_upstream_tasks(t1)) == set(wf.get_upstream_tasks(t2)) == set()
    assert set(wf.get_upstream_tasks(t3)) == {t1, t2}


def test_add(tasks):
    t1, t2, t3, t4 = tasks
    wb1 = WorkflowBuilder(tasks=[t1])
    wb1.add_task(t2, predecessors=t1)
    wb2 = WorkflowBuilder(tasks=[t3])
    wb2.add_task(t4, predecessors=t3)

    wb = wb1 + wb2

    wf = Workflow(wb)
    assert len(wf.tasks) == 4
    assert len(wf.input_tasks) == 2
    assert len(wf.output_tasks) == 2



import pytest

from pharmpy.tools.crossval import psn_crossval_results


def test_psn_crossval_results(testdata):
    res = psn_crossval_results(testdata / 'psn' / 'crossval_dir1')
    assert list(res.runs['estimation_ofv']) == pytest.approx(
        [545.863847, 512.953185, 547.843632, 579.109635]
    )
    assert list(res.runs['prediction_ofv']) == pytest.approx(
        [186.192727, 223.694832, 184.022535, 153.596550]
    )
    assert res.prediction_ofv_sum == pytest.approx(747.506644594254)



"""
.. list-table:: Options for the data module
   :widths: 25 25 50 150
   :header-rows: 1

   * - Option name
     - Default value
     - Type
     - Description
   * - ``default_dispatcher``
     - ``pharmpy.workflows.local_dask``
     - str
     - Name of default dispatcher module
   * - ``default_model_database``
     - ``pharmpy.workflows.LocalDirectoryDatabase``
     - str
     - Name of default model database class
   * - ``default_tool_database``
     - ``pharmpy.workflows.LocalDirectoryToolDatabase``
     - str
     - Name of default tool database class

"""

import importlib

import pharmpy.config as config

from .args import split_common_options
from .call import call_workflow
from .dispatchers import local_dask
from .execute import execute_workflow
from .log import Log
from .model_database import (
    LocalDirectoryDatabase,
    LocalModelDirectoryDatabase,
    ModelDatabase,
    NullModelDatabase,
)
from .model_entry import ModelEntry
from .results import ModelfitResults, Results
from .task import Task
from .tool_database import LocalDirectoryToolDatabase, NullToolDatabase, ToolDatabase
from .workflow import Workflow, WorkflowBuilder


class WorkflowConfiguration(config.Configuration):
    module = 'pharmpy.workflows'
    default_dispatcher = config.ConfigItem(
        'pharmpy.workflows.local_dask', 'Name of default dispatcher module'
    )
    default_model_database = config.ConfigItem(
        'pharmpy.workflows.LocalDirectoryDatabase', 'Name of default model database class'
    )
    default_tool_database = config.ConfigItem(
        'pharmpy.workflows.LocalDirectoryToolDatabase', 'Name of default tool database class'
    )


conf = WorkflowConfiguration()


def _importclass(name):
    a = name.split('.')
    module_name = '.'.join(a[:-1])
    if module_name == __name__:
        return globals()[a[-1]]
    else:
        module = importlib.import_module(module_name)
        return module.a[-1]


default_dispatcher = _importclass(conf.default_dispatcher)
default_model_database = _importclass(conf.default_model_database)
default_tool_database = _importclass(conf.default_tool_database)


__all__ = [
    'call_workflow',
    'default_dispatcher',
    'default_model_database',
    'default_tool_database',
    'execute_workflow',
    'split_common_options',
    'local_dask',
    'LocalDirectoryDatabase',
    'LocalModelDirectoryDatabase',
    'LocalDirectoryToolDatabase',
    'Log',
    'NullModelDatabase',
    'NullToolDatabase',
    'ModelDatabase',
    'ModelEntry',
    'ModelfitResults',
    'Results',
    'Task',
    'ToolDatabase',
    'Workflow',
    'WorkflowBuilder',
]



from dataclasses import dataclass
from typing import Generic, Literal, TypeVar

T = TypeVar('T', str, Literal[''])


class Symbol:
    pass


@dataclass(frozen=True)
class Name(Symbol, Generic[T]):
    name: T


@dataclass(frozen=True)
class Wildcard(Symbol):
    pass


@dataclass(frozen=True)
class Option(Symbol):
    option: bool



import json
import shutil
from contextlib import contextmanager
from os import stat
from pathlib import Path
from typing import Union

from pharmpy.internals.df import hash_df_fs
from pharmpy.internals.fs.lock import path_lock
from pharmpy.internals.fs.path import path_absolute
from pharmpy.model import DataInfo, Model
from pharmpy.workflows.model_entry import ModelEntry
from pharmpy.workflows.results import ModelfitResults, read_results

from .baseclass import (
    ModelSnapshot,
    ModelTransaction,
    NonTransactionalModelDatabase,
    PendingTransactionError,
    TransactionalModelDatabase,
)

DIRECTORY_PHARMPY_METADATA = '.pharmpy'
DIRECTORY_DATASETS = '.datasets'
DIRECTORY_INDEX = '.hash'
FILE_METADATA = 'metadata.json'
FILE_MODELFIT_RESULTS = 'results.json'
FILE_PENDING = 'PENDING'
FILE_LOCK = '.lock'


def get_modelfit_results(model, path):
    # FIXME: This is a workaround. The proper solution is to only read the results.json from
    # the database. For this to work roundtrip of DataFrames in json is needed.
    # This is currently broken because of rounding issue in pandas
    # Also the modelfit_results attribute will soon be removed from model objects.
    import pharmpy.model.external.nonmem as nonmem_model
    import pharmpy.tools.external.nonmem as nonmem

    if isinstance(model, nonmem_model.Model):
        res = nonmem.parse_modelfit_results(model, path)
    else:
        import pharmpy.model.external.nlmixr as nlmixr_model
        import pharmpy.tools.external.nlmixr as nlmixr

        assert isinstance(model, nlmixr_model.Model)
        res = nlmixr.parse_modelfit_results(model, path)

    model = model.replace(modelfit_results=res)
    return model


def create_model_entry(model, modelfit_results):
    # FIXME: This function is to avoid duplication of this logic, this can be removed once
    #  parent_model has been moved from Model and log has been moved from modelfit_results
    #  and each database implementation has methods for retrieving these
    if not isinstance(modelfit_results, ModelfitResults):
        modelfit_results = None
        log = None
    else:
        log = modelfit_results.log

    parent_model = model.parent_model

    return ModelEntry(model=model, modelfit_results=modelfit_results, parent=parent_model, log=log)


class LocalDirectoryDatabase(NonTransactionalModelDatabase):
    """ModelDatabase implementation for single local directory

    All files will be stored in the same directory. It is assumed that
    all files connected to a model are named modelname + extension. This means that
    care must be taken to keep filenames unique. Clashing filenames will
    be overwritten. It is recommended to use the LocalModelDirectoryDatabase
    instead.

    Parameters
    ----------
    path : str or Path
        Path to the database directory. Will be created if it does not exist.
    file_extension : str
        File extension to use for model files.
    """

    def __init__(self, path='.', file_extension='.mod'):
        path = Path(path)
        path.mkdir(parents=True, exist_ok=True)
        self.path = path_absolute(path)
        self.file_extension = file_extension
        self.ignored_names = frozenset(('stdout', 'stderr', 'nonmem.json', 'nlmixr.json'))

    def store_model(self, model):
        pass

    def store_local_file(self, model, path, new_filename=None):
        path_object = Path(path)
        if path_object.name not in self.ignored_names and path_object.is_file():
            dest_path = self.path
            if new_filename:
                dest_path = self.path / new_filename
            shutil.copy2(path, dest_path)

    def retrieve_local_files(self, name, destination_path):
        # Retrieve all files stored for one model
        files = self.path.glob(f'{name}.*')
        for f in files:
            shutil.copy2(f, destination_path)

    def retrieve_file(self, name, filename):
        # Return path to file
        path = self.path / filename
        if path.is_file() and stat(path).st_size > 0:
            return path
        else:
            raise FileNotFoundError(f"Cannot retrieve {filename} for {name}")

    def retrieve_model(self, name):
        filename = name + self.file_extension
        path = self.path / filename
        from pharmpy.model import Model

        try:
            model = Model.parse_model(path)
        except FileNotFoundError:
            raise KeyError('Model cannot be found in database')
        model = get_modelfit_results(model, self.path)
        return model

    def retrieve_modelfit_results(self, name):
        return self.retrieve_model(name).modelfit_results

    def retrieve_model_entry(self, name):
        model = self.retrieve_model(name)
        modelfit_results = self.retrieve_modelfit_results(name)
        return create_model_entry(model, modelfit_results)

    def store_metadata(self, model, metadata):
        pass

    def store_modelfit_results(self, model):
        pass

    def store_model_entry(self, model_entry):
        pass

    def __repr__(self):
        return f"LocalDirectoryDatabase({self.path})"


class LocalModelDirectoryDatabase(TransactionalModelDatabase):
    """ModelDatabase implementation for a local directory structure

    Files will be stored in separate subdirectories named after each model.
    There are no restrictions on names of the files so models can have the same
    name of some connected file without creating a name clash.

    Parameters
    ----------
    path : str or Path
        Path to the base database directory. Will be created if it does not exist.
    file_extension : str
        File extension to use for model files.
    """

    def __init__(self, path: Union[str, Path] = '.', file_extension='.mod'):
        path = Path(path)
        path.mkdir(parents=True, exist_ok=True)
        self.path = path_absolute(path)
        self.file_extension = file_extension

    def _read_lock(self):
        # NOTE: Obtain shared (blocking) lock on the entire database
        path = self.path / FILE_LOCK
        path.touch(exist_ok=True)
        return path_lock(str(path), shared=True)

    def _write_lock(self):
        # NOTE: Obtain exclusive (blocking) lock on the entire database
        path = self.path / FILE_LOCK
        path.touch(exist_ok=True)
        return path_lock(str(path), shared=False)

    @contextmanager
    def snapshot(self, model_name: str):
        model_path = self.path / model_name
        destination = model_path / DIRECTORY_PHARMPY_METADATA
        destination.mkdir(parents=True, exist_ok=True)
        with self._read_lock():
            # NOTE: Check that no pending transaction exists
            path = destination / FILE_PENDING
            if path.exists():
                # TODO: Finish pending transaction from journal if possible
                raise PendingTransactionError()

            yield LocalModelDirectoryDatabaseSnapshot(self, model_name)

    @contextmanager
    def transaction(self, model_or_model_entry: Union[Model, ModelEntry]):
        if isinstance(model_or_model_entry, ModelEntry):
            model_name = model_or_model_entry.model.name
        else:
            model_name = model_or_model_entry.name
        model_path = self.path / model_name
        destination = model_path / DIRECTORY_PHARMPY_METADATA
        destination.mkdir(parents=True, exist_ok=True)
        with self._write_lock():
            # NOTE: Mark state as pending
            path = destination / FILE_PENDING
            try:
                path.touch(exist_ok=False)
            except FileExistsError:
                # TODO: Finish pending transaction from journal if possible
                raise PendingTransactionError()

            yield LocalModelDirectoryDatabaseTransaction(self, model_or_model_entry)

            # NOTE: Commit transaction (only if no exception was raised)
            path.unlink()

    def list_models(self):
        model_dir_names = [p.name for p in self.path.glob('*') if not p.name.startswith('.')]
        return sorted(model_dir_names)

    def __repr__(self):
        return f"LocalModelDirectoryDatabase({self.path})"


class LocalModelDirectoryDatabaseTransaction(ModelTransaction):
    def __init__(
        self, database: LocalModelDirectoryDatabase, model_or_model_entry: Union[Model, ModelEntry]
    ):
        self.db = database
        if isinstance(model_or_model_entry, ModelEntry):
            self.model = model_or_model_entry.model
            self.model_entry = model_or_model_entry
        else:
            self.model = model_or_model_entry
            self.model_entry = None

    def store_model(self):
        from pharmpy.modeling import read_dataset_from_datainfo, write_csv, write_model

        model = self.model
        datasets_path = self.db.path / DIRECTORY_DATASETS

        # NOTE: Get the hash of the dataset and list filenames with contents
        # matching this hash only
        h = hash_df_fs(model.dataset)
        h_dir = datasets_path / DIRECTORY_INDEX / h
        h_dir.mkdir(parents=True, exist_ok=True)
        for hpath in h_dir.iterdir():
            # NOTE: This variable holds a string similar to "run1.csv"
            matching_model_filename = hpath.name
            data_path = datasets_path / matching_model_filename
            dipath = data_path.with_suffix('.datainfo')
            # TODO: Maybe catch FileNotFoundError and similar here (pass)
            curdi = DataInfo.read_json(dipath)
            # NOTE: Paths are not compared here
            if curdi == model.datainfo:
                df = read_dataset_from_datainfo(curdi)
                if df.equals(model.dataset):
                    # NOTE: Update datainfo path
                    datainfo = model.datainfo.replace(path=curdi.path)
                    model = model.replace(datainfo=datainfo)
                    break
        else:
            model_filename = model.name + '.csv'

            # NOTE: Create the index file at .datasets/.hash/<hash>/<model_filename>
            index_path = h_dir / model_filename
            index_path.touch()

            data_path = path_absolute(datasets_path / model_filename)
            datainfo = model.datainfo.replace(path=data_path)
            model = model.replace(datainfo=datainfo)
            model = write_csv(model, path=data_path, force=True)

            # NOTE: Write datainfo last so that we are "sure" dataset is there
            # if datainfo is there
            model.datainfo.to_json(datasets_path / (model.name + '.datainfo'))

        # NOTE: Write the model
        model_path = self.db.path / model.name
        model_path.mkdir(exist_ok=True)
        write_model(model, str(model_path / (model.name + model.filename_extension)), force=True)
        return model

    def store_local_file(self, path, new_filename=None):
        if Path(path).is_file():
            destination = self.db.path / self.model.name
            destination.mkdir(parents=True, exist_ok=True)
            if new_filename:
                destination = destination / new_filename
            shutil.copy2(path, destination)

    def store_metadata(self, metadata):
        destination = self.db.path / self.model.name / DIRECTORY_PHARMPY_METADATA
        destination.mkdir(parents=True, exist_ok=True)
        with open(destination / FILE_METADATA, 'w') as f:
            json.dump(metadata, f, indent=2)

    def store_modelfit_results(self):
        destination = self.db.path / self.model.name / DIRECTORY_PHARMPY_METADATA
        destination.mkdir(parents=True, exist_ok=True)

        if self.model_entry is not None:
            modelfit_results = self.model_entry.modelfit_results
        else:
            modelfit_results = self.model.modelfit_results

        if modelfit_results is not None:
            modelfit_results.to_json(destination / FILE_MODELFIT_RESULTS)

    def store_model_entry(self):
        if self.model_entry is None:
            raise ValueError('Transaction does not have `model_entry` attribute')
        # FIXME: Store parent
        self.store_model()
        self.store_modelfit_results()


class LocalModelDirectoryDatabaseSnapshot(ModelSnapshot):
    def __init__(self, database: LocalModelDirectoryDatabase, model_name: str):
        self.db = database
        self.name = model_name

    def retrieve_local_files(self, destination_path):
        path = self.db.path / self.name
        files = path.glob('*')
        for f in files:
            if f.is_file():
                shutil.copy2(f, destination_path)
            else:
                shutil.copytree(f, Path(destination_path) / f.stem)

    def retrieve_file(self, filename):
        # Return path to file
        path = self.db.path / self.name / filename
        if path.is_file() and stat(path).st_size > 0:
            return path
        else:
            raise FileNotFoundError(f"Cannot retrieve {filename} for {self.name}")

    def retrieve_model(self):
        extensions = ['.mod', '.ctl']
        from pharmpy.model import Model

        errors = []
        root = self.db.path / self.name
        for extension in extensions:
            filename = self.name + extension
            path = root / filename
            try:
                # NOTE: This will guess the model type
                model = Model.parse_model(path)
                break
            except FileNotFoundError as e:
                errors.append(e)
                pass
        else:
            raise KeyError(
                f'Could not find {self.name} in {self.db}.'
                f' Looked up {", ".join(map(lambda e: f"`{e.filename}`", errors))}.'
            )

        model = get_modelfit_results(model, path)
        return model

    def retrieve_modelfit_results(self):
        res = self.retrieve_model().modelfit_results
        if res is not None:
            return res

        # FIXME: The following does not work because deserialization of modelfit
        # results is not generic enough. We only use it to make the resume_tool
        # test pass.
        path = self.db.path / self.name / DIRECTORY_PHARMPY_METADATA / FILE_MODELFIT_RESULTS
        return read_results(path)

    def retrieve_model_entry(self):
        model = self.retrieve_model()
        modelfit_results = self.retrieve_modelfit_results()
        return create_model_entry(model, modelfit_results)



import numpy as np
import pytest

from pharmpy.modeling import add_allometry, add_covariate_effect
from pharmpy.modeling.lrt import best_of_many, best_of_two, cutoff, degrees_of_freedom, p_value
from pharmpy.modeling.lrt import test as lrt_test


@pytest.mark.parametrize(
    ('model_path', 'effects', 'expected', 'allow_nested'),
    [
        (
            ('nonmem', 'pheno.mod'),
            [],
            0,
            False,
        ),
        (
            ('nonmem', 'pheno.mod'),
            [('CL', 'WGT', 'exp', '*')],
            1,
            False,
        ),
        (
            ('nonmem', 'pheno.mod'),
            [('CL', 'APGR', 'cat', '*')],
            9,
            False,
        ),
        (
            ('nonmem', 'pheno_real.mod'),
            [('CL', 'WGT', 'piece_lin', '*')],
            2,
            True,
        ),
        (
            ('nonmem', 'pheno_real.mod'),
            [('CL', 'WGT', 'theta1 * (cov/median)**theta2', '*')],
            2,
            True,
        ),
        (
            ('nonmem', 'pheno_real.mod'),
            [('CL', 'WGT', '((cov/std) - median) * theta', '*')],
            1,
            True,
        ),
        (
            ('nonmem', 'pheno_real.mod'),
            [
                ('CL', 'WGT', 'exp', '+'),
                ('V', 'WGT', 'exp', '+'),
            ],
            2,
            True,
        ),
    ],
    ids=repr,
)
def test_degrees_of_freedom(
    load_model_for_test, testdata, model_path, effects, expected, allow_nested
):
    parent = load_model_for_test(testdata.joinpath(*model_path))
    child = parent

    for effect in effects:
        child = add_covariate_effect(child, *effect, allow_nested=allow_nested)

    assert degrees_of_freedom(parent, child) == expected


def test_cutoff(load_model_for_test, testdata):
    parent = load_model_for_test(testdata / 'nonmem' / 'pheno.mod')
    assert cutoff(parent, parent, 0.05) == 0

    child = add_allometry(
        parent,
        allometric_variable='WGT',
        reference_value=70,
        parameters=['CL'],
        initials=[0.7],
        lower_bounds=[0],
        upper_bounds=[2],
        fixed=True,
    )

    assert cutoff(parent, child, 0.05) == pytest.approx(3.8414588206941285)
    assert cutoff(child, parent, 0.05) == -cutoff(parent, child, 0.05)


def test_p_value(load_model_for_test, testdata):
    parent = load_model_for_test(testdata / 'nonmem' / 'pheno_abbr.mod')
    child = load_model_for_test(testdata / 'nonmem' / 'pheno_real.mod')

    alpha = 0.05
    parent_ofv = 10
    child_ofv = parent_ofv - cutoff(parent, child, alpha)

    assert p_value(parent, child, parent_ofv, child_ofv) == pytest.approx(alpha)


def test_test(load_model_for_test, testdata):
    parent = load_model_for_test(testdata / 'nonmem' / 'pheno_abbr.mod')
    child = load_model_for_test(testdata / 'nonmem' / 'pheno_real.mod')

    alpha = 0.05
    parent_ofv = 10
    child_ofv = parent_ofv - cutoff(parent, child, alpha) - 0.01

    assert lrt_test(parent, child, parent_ofv, child_ofv, alpha)


def test_best_of_two(load_model_for_test, testdata):
    parent = load_model_for_test(testdata / 'nonmem' / 'pheno_abbr.mod')
    child = load_model_for_test(testdata / 'nonmem' / 'pheno_real.mod')

    alpha = 0.05
    parent_ofv = 10
    child_ofv = parent_ofv - cutoff(parent, child, alpha) - 0.01

    assert best_of_two(parent, child, parent_ofv, child_ofv, alpha) is child


def test_best_of_many(load_model_for_test, testdata):
    parent = load_model_for_test(testdata / 'nonmem' / 'pheno_abbr.mod')
    child = load_model_for_test(testdata / 'nonmem' / 'pheno_real.mod')

    alpha = 0.05
    parent_ofv = 10
    child_ofv = parent_ofv - cutoff(parent, child, alpha) - 0.01

    assert (
        best_of_many(parent, [parent, child], parent_ofv, [parent_ofv, child_ofv], alpha) is child
    )


def test_best_of_many_nan(load_model_for_test, testdata):
    parent = load_model_for_test(testdata / 'nonmem' / 'pheno_abbr.mod')
    child = load_model_for_test(testdata / 'nonmem' / 'pheno_real.mod')

    alpha = 0.05
    parent_ofv = 10
    child_ofv = parent_ofv - cutoff(parent, child, alpha) - 0.01

    assert (
        best_of_many(parent, [child, child, child], parent_ofv, [np.nan, np.nan, np.nan], alpha)
        is parent
    )
    assert (
        best_of_many(parent, [child, child, child], parent_ofv, [np.nan, child_ofv, np.nan], alpha)
        is child
    )
    assert (
        best_of_many(
            parent, [child, child, child], parent_ofv, [np.nan, child_ofv + 0.02, np.nan], alpha
        )
        is parent
    )



from typing import Dict, List, Set, Tuple

import pharmpy.tools.modelfit as modelfit
from pharmpy.internals.expr.subs import subs
from pharmpy.internals.set.partitions import partitions
from pharmpy.internals.set.subsets import non_empty_subsets
from pharmpy.model import Model, RandomVariables
from pharmpy.modeling import (
    create_joint_distribution,
    get_omegas,
    remove_iiv,
    split_joint_distribution,
)
from pharmpy.modeling.expressions import get_rv_parameters
from pharmpy.tools.common import update_initial_estimates
from pharmpy.workflows import ModelEntry, Task, Workflow, WorkflowBuilder
from pharmpy.workflows.results import mfr


def brute_force_no_of_etas(base_model, index_offset=0, keep=None):
    wb = WorkflowBuilder(name='brute_force_no_of_etas')

    base_model = base_model.replace(description=create_description(base_model))

    iivs = base_model.random_variables.iiv
    iiv_names = iivs.names
    if keep:
        iiv_names = _remove_sublist(iiv_names, _get_eta_from_parameter(base_model, keep))

    # Remove fixed etas
    fixed_etas = _get_fixed_etas(base_model)
    iiv_names = _remove_sublist(iiv_names, fixed_etas)

    for i, to_remove in enumerate(non_empty_subsets(iiv_names), 1):
        model_name = f'iivsearch_run{i + index_offset}'
        task_candidate_entry = Task(
            'candidate_entry', create_no_of_etas_candidate_entry, model_name, to_remove
        )
        wb.add_task(task_candidate_entry)

    wf_fit = modelfit.create_fit_workflow(n=len(wb.output_tasks))
    wb.insert_workflow(wf_fit)
    return Workflow(wb)


def brute_force_block_structure(base_model, index_offset=0):
    wb = WorkflowBuilder(name='brute_force_block_structure')

    base_model = base_model.replace(description=create_description(base_model))

    iivs = base_model.random_variables.iiv
    model_no = 1 + index_offset

    fixed_etas = _get_fixed_etas(base_model)
    iiv_names = _remove_sublist(iivs.names, fixed_etas)

    for block_structure in _rv_block_structures(iiv_names):
        if _is_rv_block_structure(iivs, block_structure, fixed_etas):
            continue

        model_name = f'iivsearch_run{model_no}'
        task_candidate_entry = Task(
            'candidate_entry', create_block_structure_candidate_entry, model_name, block_structure
        )
        wb.add_task(task_candidate_entry)

        model_no += 1

    wf_fit = modelfit.create_fit_workflow(n=len(wb.output_tasks))
    wb.insert_workflow(wf_fit)
    return Workflow(wb)


def create_no_of_etas_candidate_entry(name, to_remove, model_entry):
    candidate_model = update_initial_estimates(model_entry.model, model_entry.modelfit_results)
    candidate_model = remove_iiv(candidate_model, to_remove)
    candidate_model = candidate_model.replace(
        name=name, description=create_description(candidate_model)
    )

    return ModelEntry.create(model=candidate_model, modelfit_results=None, parent=model_entry.model)


def create_block_structure_candidate_entry(name, block_structure, model_entry):
    candidate_model = update_initial_estimates(model_entry.model, model_entry.modelfit_results)
    candidate_model = create_eta_blocks(block_structure, candidate_model)
    candidate_model = candidate_model.replace(
        name=name, description=create_description(candidate_model)
    )

    return ModelEntry.create(model=candidate_model, modelfit_results=None, parent=model_entry.model)


def _rv_block_structures(etas: RandomVariables):
    # NOTE: All possible partitions of etas into block structures
    return partitions(etas)


def _is_rv_block_structure(
    etas: RandomVariables, partition: Tuple[Tuple[str, ...], ...], fixed_etas
):
    parts = set(partition)
    # Remove fixed etas from etas
    list_of_tuples = list(
        filter(
            None, list(map(lambda dist: tuple(_remove_sublist(list(dist.names), fixed_etas)), etas))
        )
    )
    return all(map(lambda dist: dist in parts, list_of_tuples))


def _create_param_dict(model: Model, dists: RandomVariables) -> Dict[str, str]:
    param_subs = {
        parameter.symbol: parameter.init for parameter in model.parameters if parameter.fix
    }
    param_dict = {}
    # FIXME: Temporary workaround, should handle IIV on eps
    symbs_before_ode = [symb.name for symb in model.statements.before_odes.free_symbols]
    for eta in dists.names:
        if subs(dists[eta].get_variance(eta), param_subs, simultaneous=True) != 0:
            # Skip etas that are before ODE
            if eta not in symbs_before_ode:
                continue
            param_dict[eta] = get_rv_parameters(model, eta)[0]
    return param_dict


def create_description(model: Model, iov: bool = False) -> str:
    if iov:
        dists = model.random_variables.iov
    else:
        dists = model.random_variables.iiv

    param_dict = _create_param_dict(model, dists)
    if len(param_dict) == 0:
        return '[]'

    blocks, same = [], []
    for dist in dists:
        rvs_names = dist.names
        param_names = [
            param_dict[name] for name in rvs_names if name not in same and name in param_dict.keys()
        ]
        if param_names:
            blocks.append(f'[{",".join(param_names)}]')

        if iov:
            same_names = []
            for name in rvs_names:
                same_names.extend(dists.get_rvs_with_same_dist(name).names)
            same.extend(same_names)

    return '+'.join(blocks)


def create_eta_blocks(partition: Tuple[Tuple[str, ...], ...], model: Model):
    for part in partition:
        if len(part) == 1:
            model = split_joint_distribution(model, part)
        else:
            model = create_joint_distribution(
                model, list(part), individual_estimates=mfr(model).individual_estimates
            )
    return model


def _get_eta_from_parameter(model: Model, parameters: List[str]) -> Set[str]:
    # returns list of eta names from parameter names
    iiv_set = set()
    iiv_names = model.random_variables.names
    for iiv_name in iiv_names:
        param = get_rv_parameters(model, iiv_name)
        if set(param).issubset(parameters) and len(param) > 0:
            iiv_set.add(iiv_name)
    return iiv_set


def _get_fixed_etas(model: Model) -> List[str]:
    fixed_omegas = get_omegas(model).fixed.names
    iivs = model.random_variables.iiv
    if len(fixed_omegas) > 0:
        fixed_etas = [
            iiv.names for iiv in iivs if str(list(iiv.variance.free_symbols)[0]) in fixed_omegas
        ]
        return [item for tup in fixed_etas for item in tup]
    else:
        return []


def _remove_sublist(list_a, list_b):
    return [x for x in list_a if x not in list_b]



"""
Generic NONMEM code record class.

"""
from __future__ import annotations

import re
from operator import neg, pos, sub, truediv
from typing import Iterator, Literal, Sequence, Set, Tuple

from pharmpy.deps import sympy, sympy_printing
from pharmpy.internals.code_generator import CodeGenerator
from pharmpy.internals.ds.ordered_set import OrderedSet
from pharmpy.internals.expr.funcs import (
    INT,
    LOG10,
    PDZ,
    PEXP,
    PHE,
    PHI,
    PLOG,
    PLOG10,
    PNG,
    PNP,
    PSQRT,
    PZR,
)
from pharmpy.internals.expr.subs import subs
from pharmpy.internals.parse import AttrTree, NoSuchRuleException
from pharmpy.internals.parse.tree import Interpreter
from pharmpy.internals.sequence.lcs import diff
from pharmpy.model import Assignment, Statement, Statements

from .parsers import CodeRecordParser
from .record import Record


class NMTranPrinter(sympy_printing.str.StrPrinter):
    _operators = {
        'not': '.NOT. ',
    }

    def __init__(self, rvs=None, trans=None, **kwargs):
        self.rvs = rvs
        self.trans = trans
        super().__init__(**kwargs)

    def _print_Add(self, expr):
        if self.rvs is None:
            return super()._print_Add(expr)
        else:
            rvs_names = set(self.rvs.names)

            trans_rvs = (
                {str(v): str(k) for k, v in self.trans.items() if str(k) in rvs_names}
                if self.trans
                else None
            )

            expr_args = expr.args
            terms_iiv_iov, terms_ruv, terms = [], [], []

            for arg in expr_args:
                arg_symbs = [s.name for s in arg.free_symbols]
                rvs_intersect = rvs_names.intersection(arg_symbs)

                if trans_rvs is not None:
                    trans_intersect = set(trans_rvs.keys()).intersection(arg_symbs)
                    rvs_intersect.update({trans_rvs[rv] for rv in trans_intersect})

                if rvs_intersect:
                    if len(rvs_intersect) == 1:
                        rv_name = list(rvs_intersect)[0]
                        variability_level = self.rvs[rv_name].level
                        if variability_level == 'RUV':
                            terms_ruv.append(arg)
                            continue
                    terms_iiv_iov.append(arg)
                else:
                    terms.append(arg)

            if not terms_iiv_iov and not terms_ruv:
                return super()._print_Add(expr)
            else:

                def arg_len(symb):
                    return len(symb.args)

                terms_iiv_iov.sort(reverse=True, key=arg_len)
                terms_ruv.sort(reverse=True, key=arg_len)
                terms += terms_iiv_iov + terms_ruv

        # Put numeric constant at the end
        for i, term in enumerate(terms):
            if term.is_Number:
                terms = terms[0:i] + terms[i + 1 :] + terms[i : i + 1]
                break

        return super()._print_Add(sympy.Add(*terms, evaluate=False), order='none')

    def _print_Float(self, expr):
        printed = sympy_printing.codeprinter.CodePrinter._print_Float(self, expr)
        return printed.upper()

    def _print_Integer(self, expr):
        return str(expr)

    def _print_Function(self, expr):
        try:
            name = expr.name
        except AttributeError:
            name = expr.__class__.__name__.upper()
        return f'{name}({super().doprint(expr.args[0])})'

    def _print_Pow(self, expr):
        if expr.exp == sympy.Rational(1, 2):
            return f"SQRT({self.doprint(expr.base)})"
        elif expr.exp == -1:
            if (
                isinstance(expr.base, sympy.Expr)
                and not isinstance(expr.base, sympy.Symbol)
                and len(expr.base.make_args(expr.base)) > 1
            ):
                return f"1/({expr.base})"
            else:
                return f"1/{expr.base}"
        else:
            return super()._print_Pow(expr)

    def _do_infix(self, expr, op):
        return super()._print(expr.args[0]) + op + super()._print(expr.args[1])

    def _print_Equality(self, expr):
        return self._do_infix(expr, ".EQ.")

    def _print_Unequality(self, expr):
        return self._do_infix(expr, ".NE.")

    def _print_Or(self, expr):
        return self._do_infix(expr, ".OR.")

    def _print_And(self, expr):
        return self._do_infix(expr, ".AND.")

    def _print_LessThan(self, expr):
        return self._do_infix(expr, ".LE.")

    def _print_StrictLessThan(self, expr):
        return self._do_infix(expr, ".LT.")

    def _print_StrictGreaterThan(self, expr):
        return self._do_infix(expr, ".GT.")

    def _print_GreaterThan(self, expr):
        return self._do_infix(expr, ".GE.")

    def _print_Not(self, expr):
        return f".NOT. ({self._print(expr.args[0])})"

    def _print_Symbol(self, expr):
        s = str(expr)
        if s != "NaN":
            s = s.upper()
        return s


def expression_to_nmtran(expr, rvs=None, trans=None):
    printer = NMTranPrinter(rvs, trans)
    expr_str = printer.doprint(expr)
    return expr_str


def nmtran_assignment_string(
    assignment: Assignment, defined_symbols: Set[sympy.Symbol], rvs, trans
):
    if isinstance(assignment.expression, sympy.Piecewise):
        return _translate_sympy_piecewise(assignment, defined_symbols, rvs, trans)
    elif re.search('sign', str(assignment.expression)):  # FIXME: Don't use re here
        return _translate_sympy_sign(assignment)
    else:
        return f'{str(assignment.symbol).upper()} = {expression_to_nmtran(assignment.expression, rvs, trans)}'


def _translate_sympy_piecewise(
    statement: Assignment, defined_symbols: Set[sympy.Symbol], rvs, trans
):
    expression = statement.expression.args
    symbol = statement.symbol
    # Did we (possibly) add the default in the piecewise with 0 or symbol?
    has_added_else = expression[-1][1] is sympy.true and (
        expression[-1][0] == symbol or (expression[-1][0] == 0 and symbol not in defined_symbols)
    )
    if has_added_else:
        expression = expression[0:-1]
    has_else = expression[-1][1] is sympy.true

    expressions, _ = zip(*expression)

    if len(expression) == 1:
        value = expression[0][0]
        condition = expression[0][1]
        condition_translated = _translate_condition(condition)

        statement_str = (
            f'IF ({condition_translated}) {symbol} = {expression_to_nmtran(value, rvs, trans)}'
        )
        return statement_str
    elif all(len(e.args) == 0 for e in expressions) and not has_else:
        return _translate_sympy_single(symbol, expression, rvs, trans)
    else:
        return _translate_sympy_block(symbol, expression, rvs, trans)


def _translate_sympy_single(symbol, expression, rvs, trans):
    statement_str = ''
    for e in expression:
        value = e[0]
        condition = e[1]

        condition_translated = _translate_condition(condition)

        statement_str += (
            f'IF ({condition_translated}) {symbol} = {expression_to_nmtran(value, rvs, trans)}\n'
        )

    return statement_str


def _translate_sympy_block(symbol, expression, rvs, trans):
    statement_str = ''
    for i, e in enumerate(expression):
        value, condition = e

        condition_translated = _translate_condition(condition)

        if i == 0:
            statement_str += f'IF ({condition_translated}) THEN\n'
        elif condition_translated == 'True':
            statement_str += 'ELSE\n'
        else:
            statement_str += f'ELSE IF ({condition_translated}) THEN\n'

        statement_str += f'    {symbol} = {expression_to_nmtran(value, rvs, trans)}\n'

    statement_str += 'END IF'
    return statement_str


def _translate_condition(c):
    fprn = NMTranPrinter()
    fortran = fprn.doprint(c).replace(' ', '')
    return fortran


def _translate_sympy_sign(s):
    args = s.expression.args

    subs_dict = {}
    for arg in args:
        if str(arg).startswith('sign'):
            sign_arg = arg.args[0]
            subs_dict[arg] = abs(sign_arg) / sign_arg

    s = s.subs(subs_dict)
    fprn = NMTranPrinter()
    fortran = fprn.doprint(s.expression)
    expr_str = f'{s.symbol} = {fortran}'
    return expr_str


class ExpressionInterpreter(Interpreter):
    def instruction_unary(self, node):
        f, x = self.visit_subtrees(node)
        return f(x)

    def instruction_infix(self, node):
        a, op, b = self.visit_subtrees(node)
        return op(a, b)

    def real_expr(self, node):
        t = self.visit_subtrees(node)
        return t[0]

    def neg_op(self, _):
        return neg

    def pos_op(self, _):
        return pos

    def add_op(self, _):
        return sympy.Add

    def sub_op(self, _):
        return sub

    def mul_op(self, _):
        return sympy.Mul

    def div_op(self, _):
        return truediv

    def pow_op(self, _):
        return sympy.Pow

    def bool_expr(self, node):
        t = self.visit_subtrees(node)
        return t[0]

    def land(self, _):
        return sympy.And

    def lor(self, _):
        return sympy.Or

    def lnot(self, _):
        return sympy.Not

    def eq(self, _):
        return sympy.Eq

    def ne(self, _):
        return sympy.Ne

    def le(self, _):
        return sympy.Le

    def lt(self, _):
        return sympy.Lt

    def ge(self, _):
        return sympy.Ge

    def gt(self, _):
        return sympy.Gt

    def img(self, node):
        fn, *parameters = self.visit_subtrees(node)
        return fn(*parameters)

    def exp(self, _):
        return sympy.exp

    def pexp(self, _):
        return PEXP

    def log(self, _):
        return sympy.log

    def plog(self, _):
        return PLOG

    def log10(self, _):
        return LOG10

    def plog10(self, _):
        return PLOG10

    def sqrt(self, _):
        return sympy.sqrt

    def psqrt(self, _):
        return PSQRT

    def sin(self, _):
        return sympy.sin

    def cos(self, _):
        return sympy.cos

    def tan(self, _):
        return sympy.tan

    def asin(self, _):
        return sympy.asin

    def acos(self, _):
        return sympy.acos

    def atan(self, _):
        return sympy.atan

    def abs(self, _):
        return sympy.Abs

    def int(self, _):
        return INT

    def loggamma(self, _):
        return sympy.loggamma

    def pdz(self, _):
        return PDZ

    def pzr(self, _):
        return PZR

    def pnp(self, _):
        return PNP

    def phe(self, _):
        return PHE

    def png(self, _):
        return PNG

    def phi(self, _):
        return PHI

    def mod(self, _):
        return sympy.Mod

    def number(self, node):
        s = str(node)
        try:
            return sympy.Integer(s)
        except ValueError:
            s = s.replace('d', 'E')  # Fortran special format
            s = s.replace('D', 'E')
            return sympy.Float(s)

    def symbol(self, node):
        t = self.visit_subtrees(node)
        return sympy.Symbol(t[0])

    def assignable(self, node):
        t = self.visit_subtrees(node)
        return sympy.Symbol(t[0])

    def parameter(self, node):
        name, *subscripts = self.visit_subtrees(node)
        return f'{name}({",".join(subscripts)})'

    def vector(self, node):
        name, *subscripts = self.visit_subtrees(node)
        return f'{name}({",".join(subscripts)})'

    def name(self, node):
        return str(node).upper()

    def index(self, node):
        return str(node)

    def array(self, node):
        name = str(node).upper()
        return 'EPS' if name == 'ERR' else name

    def matrix(self, node):
        return str(node).upper()


def _index_statements_diff(
    last_node_index: int,
    index: Sequence[Tuple[int, int, int, int]],
    it: Iterator[Tuple[Literal[-1, 0, 1], Statement]],
):
    """This function reorders and groups a diff of statements according to the
    given index mapping"""

    index_index = 0

    while True:
        try:
            op, s = next(it)
        except StopIteration:
            break

        # NOTE: We forward standalone insertions
        if op == 1:
            yield op, [s], last_node_index, last_node_index
            continue

        # NOTE: We fetch the index entry for this group of statements
        assert index_index < len(index)
        ni, nj, si, sj = index[index_index]
        index_index += 1

        operations = [op]
        statements = [s]
        expected = sj - si - 1

        # NOTE: We retrieve all statements for this index entry, as well as
        # interleaved insertion statements
        while expected > 0:
            op, s = next(it)  # NOTE: We do not expect this to raise

            operations.append(op)
            statements.append(s)
            if op != 1:
                expected -= 1

        if all(op == 0 for op in operations):
            # NOTE: If this group of statements contains no changes, we can keep
            # the associated nodes.
            yield 0, statements, ni, nj
        else:
            # NOTE: Otherwise we remove all associated nodes
            yield -1, [s for s, op in zip(statements, operations) if op != 1], ni, nj
            # NOTE: And generate new nodes for kept statements
            new_statements = [s for s, op in zip(statements, operations) if op != -1]
            if new_statements:
                yield 1, new_statements, nj, nj

        last_node_index = nj


class CodeRecord(Record):
    def __init__(self, name, raw_name, content, index=None, statements=None):
        # NOTE: self._index establishes a correspondance between self.root
        # nodes and self.statements statements. self._index consists of
        # (ni, nj, si, sj) tuples which maps the nodes
        # self.root.children[ni:nj] to the statements self.statements[si:sj]
        if index is None:
            self._index = []
        else:
            self._index = index
        if statements is not None:
            self._statements = statements
        super().__init__(name, raw_name, content)

    @property
    def statements(self):
        index, statements = _parse_tree(self.root)
        self._index = index
        self._statements = statements
        return statements

    def update_statements(self, new: Sequence[Statement], rvs=None, trans=None):
        try:
            old = self._statements
        except AttributeError:
            old = self.statements
        if new == old:
            return self
        new_children = []
        last_node_index = 0
        new_index = []
        defined_symbols: Set[sympy.Symbol] = set()  # NOTE: Set of all defined symbols so far
        si = 0  # NOTE: We keep track of progress in the "new" statements sequence

        first_statement_index = (
            self._index[0][0]  # NOTE: Start insertion just before the first statement
            if self._index
            else (
                first_verbatim  # NOTE: Start insertion just before the first verbatim
                if (first_verbatim := self.root.first_index('verbatim')) != -1
                else len(self.root.children)  # NOTE: Start insertion after all blanks
            )
        )
        for op, statements, ni, nj in _index_statements_diff(
            first_statement_index, self._index, diff(old, new)
        ):
            # NOTE: We copy interleaved non-statement nodes
            new_children.extend(self.root.children[last_node_index:ni])
            if op == 1:
                for s in statements:
                    assert isinstance(s, Assignment)
                    statement_nodes = self._statement_to_nodes(defined_symbols, s, rvs, trans)
                    # NOTE: We insert the generated nodes just before the next
                    # existing statement node
                    insert_pos = len(new_children)
                    new_index.append((insert_pos, insert_pos + len(statement_nodes), si, si + 1))
                    si += 1
                    new_children.extend(statement_nodes)
                    defined_symbols.add(s.symbol)
            elif op == 0:
                # NOTE: We keep the nodes but insert them at an updated position
                insert_pos = len(new_children)
                insert_len = nj - ni
                new_index.append((insert_pos, insert_pos + insert_len, si, si + len(statements)))
                si += len(statements)
                new_children.extend(self.root.children[ni:nj])
                for s in statements:
                    if isinstance(s, Assignment):
                        defined_symbols.add(s.symbol)
            last_node_index = nj
        # NOTE: We copy any non-statement nodes that are remaining
        new_children.extend(self.root.children[last_node_index:])
        new_root = AttrTree(self.root.rule, tuple(new_children))
        return CodeRecord(self.name, self.raw_name, new_root, index=new_index, statements=new)

    def _statement_to_nodes(self, defined_symbols: Set[sympy.Symbol], s: Assignment, rvs, trans):
        statement_str = nmtran_assignment_string(s, defined_symbols, rvs, trans) + '\n'
        node_tree = CodeRecordParser(statement_str).root
        assert node_tree is not None
        statement_nodes = list(node_tree.subtrees('statement'))
        return statement_nodes

    def from_odes(self, ode_system, extra):
        """Set statements of record given an explicit ode system
        extra statements are added to the top
        """
        odes = ode_system.eqs
        functions = [ode.lhs.args[0] for ode in odes]
        function_map = {f: sympy.Symbol(f'A({i + 1})') for i, f in enumerate(functions)}
        statements = [s.subs(function_map) for s in extra]
        for i, ode in enumerate(odes):
            # For now Piecewise signals zero-order infusions, which are handled with parameters
            ode = ode.replace(sympy.Piecewise, lambda a1, a2: 0)
            symbol = sympy.Symbol(f'DADT({i + 1})')
            expression = subs(ode.rhs, function_map, simultaneous=True)
            statements.append(Assignment(symbol, expression))
        return self.update_statements(statements)

    def update_extra_nodes(self, dvs):
        """Update AST nodes not part of the statements

        Currently the block IF for DVID is handled as extra nodes
        """
        # Want to find block_if node not in index
        if len(dvs) < 2:
            return self
        curn = 0
        found = None
        for ni, nj, _, _ in self._index:
            for n in range(curn, ni):
                node = self.root.children[n]
                if node.find('block_if') or node.find('logical_if'):
                    found = n
                    break
            curn = nj
        if curn < len(self.root.children):
            for n in range(curn, len(self.root.children)):
                node = self.root.children[n]
                if node.find('block_if') or node.find('logical_if'):
                    found = n
                    break

        if found is None:
            node = create_dvs_node(dvs)
            new_root = AttrTree(self.root.rule, self.root.children + (node,))
            rec = CodeRecord(
                self.name, self.raw_name, new_root, index=self._index, statements=self._statements
            )
            return rec
        else:
            return self


def create_dvs_node(dvs):
    """Create special dvs AST node"""
    for i, (dv, dvid) in enumerate(dvs.items()):
        cg = CodeGenerator()
        cg.add(f'IF (DVID.EQ.{dvid}) Y = {dv}')

        if i == 0:
            node = CodeRecordParser(str(cg)).root.children[0]
        elif i == len(dvs.items()) - 1:
            node = node.add_node(CodeRecordParser(str(cg) + '\n').root.children[0])
        else:
            node = node.add_node(CodeRecordParser(str(cg)).root.children[0])
    return node


def _parse_tree(tree: AttrTree):
    s = []
    new_index = []
    interpreter = ExpressionInterpreter()
    for child_index, child in enumerate(tree.children):
        if not isinstance(child, AttrTree) or child.rule != 'statement':
            continue
        for node in child.children:
            # NOTE: Why does this iterate over the children?
            # Right now it looks like it could add the same statement
            # multiple times because there is no break on a match.
            if not isinstance(node, AttrTree):
                continue
            if node.rule == 'assignment':
                symbol = interpreter.visit(node.subtree('assignable'))
                expr = interpreter.visit(node.subtree('real_expr'))
                ass = Assignment(symbol, expr)
                s.append(ass)
                new_index.append((child_index, child_index + 1, len(s) - 1, len(s)))
            elif node.rule == 'logical_if':
                logic_expr = interpreter.visit(node.subtree('bool_expr'))
                try:
                    assignment = node.subtree('assignment')
                except NoSuchRuleException:
                    pass
                else:
                    symbol = interpreter.visit(assignment.subtree('assignable'))
                    expr = interpreter.visit(assignment.subtree('real_expr'))
                    # Check if symbol was previously declared
                    else_val = (
                        symbol if any(map(lambda x: x.symbol == symbol, s)) else sympy.Integer(0)
                    )
                    pw = sympy.Piecewise((expr, logic_expr), (else_val, True))
                    ass = Assignment(symbol, pw)
                    s.append(ass)
                    new_index.append((child_index, child_index + 1, len(s) - 1, len(s)))
            elif node.rule == 'block_if':
                blocks = []  # [(logic, [(symb1, expr1), ...]), ...]
                symbols = OrderedSet()

                first_block = node.subtree('block_if_start')
                first_logic = interpreter.visit(first_block.subtree('bool_expr'))
                first_symb_exprs = []
                for ifstat in first_block.subtrees('statement'):
                    for assign_node in ifstat.subtrees('assignment'):
                        symbol = interpreter.visit(assign_node.subtree('assignable'))
                        first_symb_exprs.append(
                            (symbol, interpreter.visit(assign_node.subtree('real_expr')))
                        )
                        symbols.add(symbol)
                blocks.append((first_logic, first_symb_exprs))

                for elseif in node.subtrees('block_if_elseif'):
                    logic = interpreter.visit(elseif.subtree('bool_expr'))
                    elseif_symb_exprs = []
                    for elseifstat in elseif.subtrees('statement'):
                        for assign_node in elseifstat.subtrees('assignment'):
                            symbol = interpreter.visit(assign_node.subtree('assignable'))
                            elseif_symb_exprs.append(
                                (symbol, interpreter.visit(assign_node.subtree('real_expr')))
                            )
                            symbols.add(symbol)
                    blocks.append((logic, elseif_symb_exprs))

                else_block = node.find('block_if_else')
                if else_block:
                    assert isinstance(else_block, AttrTree)
                    else_symb_exprs = []
                    for elsestat in else_block.subtrees('statement'):
                        for assign_node in elsestat.subtrees('assignment'):
                            symbol = interpreter.visit(assign_node.subtree('assignable'))
                            else_symb_exprs.append(
                                (symbol, interpreter.visit(assign_node.subtree('real_expr')))
                            )
                            symbols.add(symbol)
                    piecewise_logic = True
                    if len(blocks) == 1 and len(blocks[0][1]) == 0:
                        # Special case for empty if
                        piecewise_logic = sympy.Not(blocks[0][0])
                    blocks.append((piecewise_logic, else_symb_exprs))
                s_block = []
                for symbol in symbols:
                    pairs = []
                    for block in blocks:
                        logic = block[0]
                        for cursymb, expr in block[1]:
                            if cursymb == symbol:
                                pairs.append((expr, logic))

                    if pairs[-1][1] is not True:
                        else_val = (
                            symbol
                            if any(map(lambda x: x.symbol == symbol, s))
                            else sympy.Integer(0)
                        )
                        pairs.append((else_val, True))

                    if len(pairs) == 1:
                        expr = pairs[0][0]
                        ass = Assignment(symbol, expr)
                    else:
                        pw = sympy.Piecewise(*pairs)
                        ass = Assignment(symbol, pw)
                    s_block.append(ass)

                s.extend(_reorder_block_statements(s_block))
                curind = (child_index, child_index + 1, len(s) - len(symbols), len(s))
                new_index.append(curind)

    return new_index, Statements(s)


def _reorder_block_statements(s):
    piecewise = [ass for ass in s if isinstance(ass.expression, sympy.Piecewise)]
    assignments = [ass for ass in s if ass not in piecewise]
    return assignments + piecewise



from functools import partial
from typing import Iterable

from pharmpy.model import Model
from pharmpy.modeling import set_direct_effect

from ..statement.feature.direct_effect import DirectEffect
from ..statement.feature.symbols import Name, Wildcard
from ..statement.statement import Statement
from .feature import Feature


def features(model: Model, statements: Iterable[Statement]) -> Iterable[Feature]:
    for statement in statements:
        if isinstance(statement, DirectEffect):
            modes = (
                [Name('LINEAR'), Name('EMAX'), Name('SIGMOID')]
                if isinstance(statement.modes, Wildcard)
                else statement.modes
            )
            for mode in modes:
                yield ('DIRECT', mode.name), partial(set_direct_effect, expr=mode.name.lower())



from __future__ import annotations

import warnings
from pathlib import Path
from typing import TYPE_CHECKING, Dict, Optional, Union

import pharmpy.modeling as modeling
from pharmpy.internals.math import nearest_positive_semidefinite
from pharmpy.model import EstimationSteps, Model, Parameters, RandomVariables
from pharmpy.model.external.nonmem.nmtran_parser import NMTranControlStream
from pharmpy.model.external.nonmem.parsing import parse_table_columns
from pharmpy.model.external.nonmem.table import ExtTable, NONMEMTableFile, PhiTable
from pharmpy.model.external.nonmem.update import create_name_map
from pharmpy.workflows.log import Log
from pharmpy.workflows.results import ModelfitResults

from .results_file import NONMEMResultsFile

if TYPE_CHECKING:
    import numpy as np
    import pandas as pd
    import sympy
else:
    from pharmpy.deps import numpy as np
    from pharmpy.deps import pandas as pd
    from pharmpy.deps import sympy


def _parse_modelfit_results(
    path: Optional[Union[str, Path]],
    control_stream: NMTranControlStream,
    name_map,
    model: Model,
    subproblem: Optional[int] = None,
):
    # Path to model file or results file
    if path is None:
        return None

    path = Path(path)

    name = model.name
    description = model.description
    estimation_steps = model.estimation_steps
    parameters = model.parameters
    etas = model.random_variables.etas

    log = Log()
    try:
        try:
            ext_path = path.with_suffix('.ext')
            ext_tables = NONMEMTableFile(ext_path)
        except ValueError:
            log = log.log_error(f"Broken ext-file {path.with_suffix('.ext')}")
            return ModelfitResults(
                name=name,
                description=description,
                minimization_successful=False,
                ofv=np.nan,
                parameter_estimates=_create_failed_parameter_estimates(model.parameters),
                log=log,
            )

        for table in ext_tables:
            try:
                table.data_frame
            except ValueError:
                log = log.log_error(
                    f"Broken table in ext-file {path.with_suffix('.ext')}, "
                    f"table no. {table.number}"
                )
    except (FileNotFoundError, OSError):
        return None

    (
        table_numbers,
        final_ofv,
        ofv_iterations,
        final_pe,
        sdcorr,
        pe_iterations,
        ses,
        ses_sdcorr,
    ) = _parse_ext(control_stream, name_map, ext_tables, subproblem, parameters)

    table_df = _parse_tables(
        path, control_stream, netas=len(model.random_variables.etas.names)
    )  # $TABLEs
    residuals = _parse_residuals(table_df)
    predictions = _parse_predictions(table_df)
    iofv, ie, iec = _parse_phi(path, control_stream, name_map, etas, model, final_pe, subproblem)
    gradients_iterations, final_zero_gradient, gradients = _parse_grd(
        path, control_stream, name_map, parameters, subproblem
    )
    rse = _calculate_relative_standard_errors(final_pe, ses)
    (
        runtime_total,
        log_likelihood,
        covstatus,
        covstatus_table_number,
        minimization_successful,
        function_evaluations,
        significant_digits,
        termination_cause,
        estimation_runtime,
        estimate_near_boundary,
        log,
        est_table_numbers,
    ) = _parse_lst(len(estimation_steps), path, table_numbers, log)

    if est_table_numbers:
        eststeps = est_table_numbers
    else:
        eststeps = list(range(1, len(estimation_steps) + 1))
    last_est_ind = _get_last_est(estimation_steps)
    minsucc_iters = pd.Series(
        minimization_successful, index=eststeps, name='minimization_successful'
    )
    esttime_iters = pd.Series(estimation_runtime, index=eststeps, name='estimation_runtime')
    funcevals_iters = pd.Series(function_evaluations, index=eststeps, name='function_evaluations')
    termcause_iters = pd.Series(termination_cause, index=eststeps, name='termination_cause')
    sigdigs_iters = pd.Series(significant_digits, index=eststeps, name='significant_digits')

    if covstatus and ses is not None:
        cov = _parse_matrix(path.with_suffix(".cov"), control_stream, name_map, table_numbers)
        cor = _parse_matrix(path.with_suffix(".cor"), control_stream, name_map, table_numbers)
        if cor is not None:
            np.fill_diagonal(cor.values, 1)
        coi = _parse_matrix(path.with_suffix(".coi"), control_stream, name_map, table_numbers)
    else:
        cov, cor, coi = None, None, None

    cov, cor, coi, ses = calculate_cov_cor_coi_ses(cov, cor, coi, ses)
    if cov is not None:
        cov = nearest_positive_semidefinite(cov)

    evaluation = _parse_evaluation(estimation_steps)

    if (
        not model.estimation_steps
        or model.estimation_steps[-1].parameter_uncertainty_method is None
    ):
        covstep_successful = None
    elif covstatus:
        covstep_successful = True
    else:
        covstep_successful = False

    warnings = []
    if any(estimate_near_boundary):
        warnings.append('estimate_near_boundary')
    if final_zero_gradient:
        warnings.append('final_zero_gradient')

    res = ModelfitResults(
        name=name,
        description=description,
        minimization_successful=minimization_successful[last_est_ind],
        minimization_successful_iterations=minsucc_iters,
        estimation_runtime=estimation_runtime[last_est_ind],
        estimation_runtime_iterations=esttime_iters,
        function_evaluations=function_evaluations[last_est_ind],
        function_evaluations_iterations=funcevals_iters,
        termination_cause=termination_cause[last_est_ind],
        termination_cause_iterations=termcause_iters,
        significant_digits=significant_digits[-1],
        significant_digits_iterations=sigdigs_iters,
        relative_standard_errors=rse,
        individual_estimates=ie,
        individual_estimates_covariance=iec,
        runtime_total=runtime_total,
        log_likelihood=log_likelihood,
        covariance_matrix=cov,
        correlation_matrix=cor,
        precision_matrix=coi,
        standard_errors=ses,
        standard_errors_sdcorr=ses_sdcorr,
        individual_ofv=iofv,
        parameter_estimates=final_pe,
        parameter_estimates_sdcorr=sdcorr,
        parameter_estimates_iterations=pe_iterations,
        ofv=final_ofv,
        ofv_iterations=ofv_iterations,
        predictions=predictions,
        residuals=residuals,
        evaluation=evaluation,
        log=log,
        covstep_successful=covstep_successful,
        gradients=gradients,
        gradients_iterations=gradients_iterations,
        warnings=warnings,
    )
    return res


def calculate_cov_cor_coi_ses(cov, cor, coi, ses):
    if cov is None:
        if cor is not None:
            cov = modeling.calculate_cov_from_corrse(cor, ses)
        elif coi is not None:
            cov = modeling.calculate_cov_from_prec(coi)
    if cor is None:
        if cov is not None:
            cor = modeling.calculate_corr_from_cov(cov)
        elif coi is not None:
            cor = modeling.calculate_corr_from_prec(coi)
    if coi is None:
        if cov is not None:
            coi = modeling.calculate_prec_from_cov(cov)
        elif cor is not None:
            coi = modeling.calculate_prec_from_corrse(cor, ses)
    if ses is None:
        if cov is not None:
            ses = modeling.calculate_se_from_cov(cov)
        elif coi is not None:
            ses = modeling.calculate_se_from_prec(coi)
    return cov, cor, coi, ses


def _parse_matrix(
    path: Path,
    control_stream: NMTranControlStream,
    name_map,
    table_numbers,
):
    try:
        tables = NONMEMTableFile(path)
    except OSError:
        return None

    last_table = tables.table_no(table_numbers[-1])
    assert last_table is not None
    df = last_table.data_frame
    if df is not None:
        df = df.rename(index=name_map)
        df.columns = df.index
    return df


def _empty_lst_results(n: int, log):
    false_vec = [False] * n
    nan_vec = [np.nan] * n
    none_vec = [None] * n
    return (
        None,
        np.nan,
        False,
        None,
        false_vec,
        nan_vec,
        nan_vec,
        none_vec,
        nan_vec,
        false_vec,
        log,
        nan_vec,
    )


def _parse_lst(n: int, path: Path, table_numbers, log: Log):
    try:
        rfile = NONMEMResultsFile(path.with_suffix('.lst'), log=log)
    except OSError:
        return _empty_lst_results(n, log)

    if not table_numbers:
        return _empty_lst_results(n, log)

    runtime_total = rfile.runtime_total

    covstatus_table_number = table_numbers[-1]

    try:
        for table_number in reversed(table_numbers):
            if "OPTIMALITY" not in rfile.table[table_number]['METH']:
                log_likelihood_table_number = table_number
                break
    except (KeyError, FileNotFoundError):
        log_likelihood_table_number = None

    try:
        log_likelihood = rfile.table[log_likelihood_table_number]['ofv_with_constant']
    except (KeyError, FileNotFoundError):
        log_likelihood = np.nan

    covstatus = rfile.covariance_status(covstatus_table_number)['covariance_step_ok']

    if log_likelihood_table_number is not None:
        est_table_numbers = [
            table_number
            for table_number in table_numbers
            if table_number <= log_likelihood_table_number
        ]
    else:
        est_table_numbers = [table_number for table_number in table_numbers]

    (
        minimization_successful,
        function_evaluations,
        significant_digits,
        termination_cause,
        estimation_runtime,
        estimate_near_boundary,
    ) = parse_estimation_status(rfile, est_table_numbers)

    return (
        runtime_total,
        log_likelihood,
        covstatus,
        covstatus_table_number,
        minimization_successful,
        function_evaluations,
        significant_digits,
        termination_cause,
        estimation_runtime,
        estimate_near_boundary,
        rfile.log,
        est_table_numbers,
    )


def parse_estimation_status(results_file, table_numbers):
    minimization_successful = []
    function_evaluations = []
    significant_digits = []
    termination_cause = []
    estimation_runtime = []
    estimate_near_boundary = []
    for tabno in table_numbers:
        if results_file is not None:
            estimation_status = results_file.estimation_status(tabno)
        else:
            estimation_status = NONMEMResultsFile.unknown_termination()
        minimization_successful.append(estimation_status['minimization_successful'])
        function_evaluations.append(estimation_status['function_evaluations'])
        significant_digits.append(estimation_status['significant_digits'])
        estimate_near_boundary.append(estimation_status['estimate_near_boundary'])
        if estimation_status['maxevals_exceeded'] is True:
            tc = 'maxevals_exceeded'
        elif estimation_status['rounding_errors'] is True:
            tc = 'rounding_errors'
        else:
            tc = None
        termination_cause.append(tc)
        try:
            er = results_file.table[tabno]['estimation_runtime']
        except (KeyError, FileNotFoundError):
            er = np.nan
        estimation_runtime.append(er)

    return (
        minimization_successful,
        function_evaluations,
        significant_digits,
        termination_cause,
        estimation_runtime,
        estimate_near_boundary,
    )


def _get_last_est(estimation_steps: EstimationSteps):
    # Find last estimation
    for i in range(len(estimation_steps) - 1, -1, -1):
        step = estimation_steps[i]
        if not step.evaluation:
            return i
    # If all steps were evaluation the last evaluation step is relevant
    return len(estimation_steps) - 1


def _parse_phi(
    path: Path,
    control_stream: NMTranControlStream,
    name_map,
    etas: RandomVariables,
    model,
    pe,
    subproblem=None,
):
    try:
        phi_tables = NONMEMTableFile(path.with_suffix('.phi'))
    except FileNotFoundError:
        return None, None, None
    if subproblem is None:
        table = None

        for t in reversed(phi_tables.tables):
            if t.design_optimality is None:
                table = t
                break
    else:
        table = phi_tables.tables[subproblem - 1]

    if table is None:
        return None, None, None

    assert isinstance(table, PhiTable)

    eta_names = set(name_map.values())
    rv_names = list(filter(eta_names.__contains__, etas.names))
    try:
        individual_ofv = table.iofv
        prefix, individual_estimates = _parse_individual_estimates(model, pe, table, rv_names)
        ids, eta_col_names, matrix_array = table.etc_data()
        index = {name_map[x]: i for i, x in enumerate(eta_col_names)}
        indices = tuple(map(index.__getitem__, rv_names))
        selector = np.ix_(indices, indices)
        etc_frames = [
            pd.DataFrame(matrix[selector], columns=rv_names, index=rv_names)
            for matrix in matrix_array
        ]
        covs = pd.Series(etc_frames, index=ids, dtype='object')
        return individual_ofv, individual_estimates, covs
    except KeyError:
        return None, None, None


def _parse_individual_estimates(model, pe, table, rv_names):
    df = table.etas
    prefix = df.columns[0][0:3]  # PHI or ETA
    d = {f"{prefix}({i})": name for i, name in enumerate(rv_names, start=1)}
    df = df.rename(columns=d)  # [rv_names] needed?
    if prefix == "PHI":
        for i in range(1, len(rv_names) + 1):
            mu = sympy.Symbol(f"MU_{i}")
            expr = model.statements.before_odes.full_expression(mu)
            if expr != mu:  # MU is defined in model code
                value = expr.subs(dict(pe)).subs(model.parameters.inits)
                try:
                    value = float(value)
                except TypeError:
                    raise ValueError("MU depends on something else than population parameters")
                df.iloc[:, i - 1] -= value
    return prefix, df


def _parse_grd(
    path: Path,
    control_stream: NMTranControlStream,
    name_map,
    parameters: Parameters,
    subproblem=None,
):
    try:
        grd_tables = NONMEMTableFile(path.with_suffix('.grd'))
    except FileNotFoundError:
        return None, None, None
    if subproblem is None:
        table = grd_tables.tables[-1]
    else:
        table = grd_tables.tables[subproblem - 1]

    if table is None:
        return None, None, None

    gradients_table = table.data_frame
    old_col_names = table.data_frame.columns.to_list()[1::]
    param_names = [name for name in list(name_map.values()) if name in parameters.names]
    new_col_names = {old_col_names[i]: param_names[i] for i in range(len(old_col_names))}
    gradients_table = gradients_table.rename(columns=new_col_names)
    last_row = gradients_table.tail(1)
    last_row = last_row.drop(columns=['ITERATION'])
    last_row = last_row.squeeze(axis=0).rename('gradients')
    final_zero_gradient = (last_row == 0).any() or last_row.isnull().any()
    return gradients_table, final_zero_gradient, last_row


def _parse_tables(path: Path, control_stream: NMTranControlStream, netas) -> pd.DataFrame:
    """Parse $TABLE and table files into one large dataframe of useful columns"""
    interesting_columns = {
        'ID',
        'TIME',
        'PRED',
        'CIPREDI',
        'CPRED',
        'IPRED',
        'RES',
        'WRES',
        'CWRES',
        'MDV',
    }

    table_recs = control_stream.get_records('TABLE')
    colnames_list = parse_table_columns(control_stream, netas)
    found = set()
    df = pd.DataFrame()
    for table_rec, colnames in zip(table_recs, colnames_list):
        columns_in_table = []
        colnames_in_table = []
        for i, name in enumerate(colnames):
            if name in interesting_columns and name not in found:
                found.add(name)
                colnames_in_table.append(name)
                columns_in_table.append(i)

        noheader = table_rec.has_option("NOHEADER")
        notitle = table_rec.has_option("NOTITLE") or noheader
        nolabel = table_rec.has_option("NOLABEL") or noheader
        table_path = path.parent / table_rec.path
        try:
            table_file = NONMEMTableFile(table_path, notitle=notitle, nolabel=nolabel)
        except IOError:
            continue
        table = table_file.tables[0]

        df[colnames_in_table] = table.data_frame.iloc[:, columns_in_table]

    if 'ID' in df.columns:
        df['ID'] = df['ID'].convert_dtypes()
    return df


def _extract_from_df(df: pd.DataFrame, mandatory, optional):
    # Extract all mandatory and at least one optional column from df
    columns = set(df.columns)
    if not (set(mandatory) <= columns):
        return None

    found_optionals = [col for col in optional if col in columns]
    if not found_optionals:
        return None
    return df[mandatory + found_optionals]


def _parse_residuals(df: pd.DataFrame):
    index_cols = ['ID', 'TIME']
    cols = ['RES', 'WRES', 'CWRES']
    df = _extract_from_df(df, index_cols, cols)
    if df is not None:
        df.set_index(['ID', 'TIME'], inplace=True)
        df = df.loc[(df != 0).any(axis=1)]  # Simple way of removing non-observations
    return df


def _parse_predictions(df: pd.DataFrame):
    index_cols = ['ID', 'TIME']
    cols = ['PRED', 'CIPREDI', 'CPRED', 'IPRED']
    df = _extract_from_df(df, index_cols, cols)
    if df is not None:
        df.set_index(['ID', 'TIME'], inplace=True)
    return df


def _create_failed_ofv_iterations(n: int):
    steps = list(range(1, n + 1))
    iterations = [0] * n
    ofvs = [np.nan] * n
    ofv_iterations = create_ofv_iterations_series(ofvs, steps, iterations)
    return ofv_iterations


def create_ofv_iterations_series(ofv, steps, iterations):
    step_series = pd.Series(steps, dtype='int32', name='steps')
    iteration_series = pd.Series(iterations, dtype='int32', name='iteration')
    with warnings.catch_warnings():
        # Needed because pandas 2.1.1 uses the _data attribute that it
        # also has deprecated
        warnings.simplefilter("ignore")
        ofv_iterations = pd.Series(
            ofv, name='OFV', dtype='float64', index=[step_series, iteration_series]
        )
    return ofv_iterations


def _create_failed_parameter_estimates(parameters: Parameters):
    return pd.Series(np.nan, name='estimates', index=list(parameters.nonfixed.inits.keys()))


def _parse_ext(
    control_stream: NMTranControlStream,
    name_map,
    ext_tables: NONMEMTableFile,
    subproblem: Optional[int],
    parameters: Parameters,
):
    table_numbers = _parse_table_numbers(ext_tables, subproblem)

    final_ofv, ofv_iterations = _parse_ofv(ext_tables, subproblem)
    final_pe, sdcorr, pe_iterations = _parse_parameter_estimates(
        control_stream, name_map, ext_tables, subproblem, parameters
    )
    ses, ses_sdcorr = _parse_standard_errors(
        control_stream, name_map, ext_tables, parameters, final_pe
    )
    return (
        table_numbers,
        final_ofv,
        ofv_iterations,
        final_pe,
        sdcorr,
        pe_iterations,
        ses,
        ses_sdcorr,
    )


def _parse_table_numbers(ext_tables: NONMEMTableFile, subproblem: Optional[int]):
    table_numbers = []
    for table in ext_tables.tables:
        if subproblem and table.subproblem != subproblem:
            continue
        table_numbers.append(table.number)
    return table_numbers


def _parse_ofv(ext_tables: NONMEMTableFile, subproblem: Optional[int]):
    step = []
    iteration = []
    ofv = []
    final_table = None
    for table_number, table in enumerate(ext_tables.tables, start=1):
        if (subproblem and table.subproblem != subproblem) or table.design_optimality is not None:
            continue
        df = _get_iter_df(table.data_frame)
        n = len(df)
        step += [table_number] * n
        iteration += list(df['ITERATION'])
        ofv += list(df['OBJ'])
        final_table = table

    assert isinstance(final_table, ExtTable)
    if np.isnan(ofv[-1]):
        final_ofv = ofv[-1]
    else:
        final_ofv = final_table.final_ofv
    ofv_iterations = create_ofv_iterations_series(ofv, step, iteration)
    return final_ofv, ofv_iterations


def _get_iter_df(df):
    final_iter = -(10**9)
    iters = df['ITERATION']
    if 0 not in iters.values and final_iter in iters.values:
        df = df[iters == -(10**9)]
        df.at[0, 'ITERATION'] = 0
    else:
        try:
            final_iter_ofv = df[iters == final_iter].iloc[-1].loc['OBJ']
        except IndexError:
            final_iter_ofv = np.nan
        last_iter_ofv = df[iters >= 0].iloc[-1].loc['OBJ']
        if final_iter_ofv != last_iter_ofv:
            new_iter_no = int(df[iters >= 0].iloc[-1].loc['ITERATION']) + 1
            if final_iter in iters.values:
                idx = df[iters == final_iter].index.values[0]
                df.at[idx, 'ITERATION'] = new_iter_no
                new_iter_no += 1
            nan_row_dict = {
                colname: (new_iter_no if colname == 'ITERATION' else np.nan)
                for colname in df.columns
            }
            nan_row = pd.DataFrame(nan_row_dict, index=[len(df)])
            df = pd.concat([df, nan_row])
        df = df[df['ITERATION'] >= 0]
    return df


def _calculate_relative_standard_errors(pe, se):
    if pe is None or se is None:
        ser = None
    else:
        ser = se / pe
        ser.name = 'RSE'
    return ser


def _parse_parameter_estimates(
    control_stream: NMTranControlStream,
    name_map,
    ext_tables: NONMEMTableFile,
    subproblem: Optional[int],
    parameters: Parameters,
):
    pe = pd.DataFrame()
    fixed_param_names = []
    final_table = None
    fix = None
    for table_number, table in enumerate(ext_tables.tables, start=1):
        if (subproblem and table.subproblem != subproblem) or table.design_optimality is not None:
            continue
        df = _get_iter_df(table.data_frame)
        assert isinstance(table, ExtTable)
        fix = _get_fixed_parameters(table, parameters, name_map)
        fixed_param_names = [name for name in list(df.columns)[1:-1] if fix[name]]
        df = df.drop(fixed_param_names + ['OBJ'], axis=1)
        df['step'] = table_number
        df = df.rename(columns=name_map)
        pe = pd.concat([pe, df])
        final_table = table

    assert fix is not None
    assert final_table is not None
    if pe.iloc[-1].drop(['ITERATION', 'step']).isnull().all():
        final_pe = pe.iloc[-1].drop(['ITERATION', 'step']).rename('estimates')
    else:
        final_pe = final_table.final_parameter_estimates
        final_pe = final_pe.drop(fixed_param_names)
        final_pe = final_pe.rename(index=name_map)
    pe = pe.rename(columns={'ITERATION': 'iteration'}).set_index(['step', 'iteration'])

    try:
        sdcorr = final_table.omega_sigma_stdcorr[~fix]
    except KeyError:
        sdcorr_ests = pd.Series(np.nan, index=pe.index)
    else:
        sdcorr = sdcorr.rename(index=name_map)
        sdcorr_ests = final_pe.copy()
        sdcorr_ests.update(sdcorr)
    return final_pe, sdcorr_ests, pe


def _parse_standard_errors(
    control_stream: NMTranControlStream,
    name_map,
    ext_tables: NONMEMTableFile,
    parameters: Parameters,
    pe: pd.Series,
):
    table = ext_tables.tables[-1]
    assert isinstance(table, ExtTable)
    try:
        ses = table.standard_errors
    except KeyError:
        ses = pd.Series(np.nan, index=pe.index, name="SE")
        sdcorr_ses = pd.Series(np.nan, index=pe.index, name="SE_sdcorr")
        return ses, sdcorr_ses

    fix = _get_fixed_parameters(table, parameters, name_map)
    ses = ses[~fix]
    sdcorr = table.omega_sigma_se_stdcorr[~fix]
    ses = ses.rename(index=name_map)
    sdcorr = sdcorr.rename(index=name_map)
    sdcorr_ses = ses.copy()
    sdcorr_ses.update(sdcorr)
    sdcorr_ses = sdcorr_ses.rename(index=name_map)
    return ses, sdcorr_ses


def _parse_evaluation(estimation_steps: EstimationSteps):
    index = list(range(1, len(estimation_steps) + 1))
    evaluation = [est.evaluation for est in estimation_steps]
    return pd.Series(evaluation, index=index, name='evaluation', dtype='float64')


def _get_fixed_parameters(table: ExtTable, parameters: Parameters, pe_translation: Dict):
    try:
        return table.fixed
    except KeyError:
        # NM 7.2 does not have row -1000000006 indicating FIXED status
        ests = table.final_parameter_estimates
        fixed = pd.Series(parameters.fix)
        # NOTE: Parameters in result file haven't been renamed yet
        fixed = fixed.rename({value: key for key, value in pe_translation.items()})
        return pd.concat([fixed, pd.Series(True, index=ests.index.difference(fixed.index))])


def simfit_results(model, model_path):
    """Read in modelfit results from a simulation/estimation model"""
    nsubs = model.internals.control_stream.get_records('SIMULATION')[0].nsubs
    results = []
    for i in range(1, nsubs + 1):
        res = parse_modelfit_results(model, model_path, subproblem=i)
        results.append(res)
    return results


# def parse_ext(model, path, subproblem):
#     try:
#         ext_tables = NONMEMTableFile(path.with_suffix('.ext'))
#     except ValueError:
#         failed_pe = _create_failed_parameter_estimates(model.parameters)
#         n = len(model.estimation_steps)
#         df = pd.concat([failed_pe] * n, axis=1).T
#         df['step'] = range(1, n + 1)
#         df['iteration'] = 0
#         df = df.set_index(['step', 'iteration'])
#         return (
#             [],
#             np.nan,
#             _create_failed_ofv_iterations(len(model.estimation_steps)),
#             failed_pe,
#             failed_pe,
#             df,
#             None,
#             None,
#         )
#     return _parse_ext(model.internals.control_stream, ext_tables, subproblem, model.parameters)


def parse_modelfit_results(
    model, path: Optional[Union[str, Path]], subproblem: Optional[int] = None
):
    name_map = create_name_map(model)
    name_map = {value: key for key, value in name_map.items()}
    res = _parse_modelfit_results(
        path,
        model.internals.control_stream,
        name_map,
        model,
        subproblem=subproblem,
    )
    return res
